#
# $Copyright (c) 2017 Veritas Technologies LLC.  All rights reserved. Veritas
# and the Veritas Logo are trademarks or registered trademarks of Veritas
# Technologies LLC or its affiliates in the U.S. and other countries. Other 
# names may be trademarks of their respective owners.
# 
# The Licensed Software and Documentation are deemed to be commercial computer
# software as defined in FAR 12.212 and subject to restricted rights as defined
# in FAR Section 52.227-19 "Commercial Computer Software - Restricted Rights" and
# DFARS 227.7202, et seq. "Commercial Computer Software and Commercial Computer
# Software Documentation," as applicable, and any successor regulations, whether
# delivered by Veritas as on premises or hosted services. Any use, modification,
# reproduction release, performance, display or disclosure of the Licensed Software
# and Documentation by the U.S. Government shall be solely in accordance with the
# terms of this Agreement. $

# $Revision: 1.12.5.63 $
# $Source: /project/arrayrep/src/srdf/unix/SRDFAgent.pm,v $
#

package SRDFAgent;
use ArrayRep;
@ISA = ("ArrayRep");

use strict;
use vars qw($VERSION @ISA @EXPORT @EXPORT_OK %default);
use POSIX qw(uname);
use File::Basename;
use ag_i18n_inc;
my $vcs_home=$VCSHOME;
my $vcs_log=$VCSLOG;
my $USERLANG=$ENV{"LANG"};
# supported microcode version
use constant SUPPORTED_MCODE_VERSION => 5568;
my ($OS, $DEVNULL) = ((uname())[0], "/dev/null");

my $VXDISK = "/usr/sbin/vxdisk";
my $powermt = '/sbin/powermt';
my $blockdev = '/sbin/blockdev';
my $native_multipathing = '/sbin/multipath';
my $VXPATH = "/dev/vx/rdmp";

my ($symrdf, $symdg, $sympd, $symcfg, $symcg, $symdev, $symcli, $syminq);

if ($OS eq "Windows NT") {
    $DEVNULL = "nul";
}

@EXPORT = qw(
);
$VERSION = '1.00';

require 5.000;

%default = (
        resname => '',
        name    => '',
        home    => '',
        mode    => '',
        state   => '',
        role    => '',
        devs    => 0,
        invalid => 0,
        devfotime => 0,
        iscompositegroup => 0,
        dcv_mode => '',
        consistency => '',
        numofRDFGs => 0,
        onlinetimeout => 0,
        swaproles => 0,
	swapcapable=>0,
        resource_type => 'SRDF',
        ARRAYDEVS => {},
        start_time => '',
        RDFG => {},
        LSymmetrix => {},
        RSymmetrix => {},
        type => '',
        concurrent => -1,
        preferredrdfg => '',
        rdfgopt => '',
        is_vmware => 0,
	grp_type => '',
	dev_state => '',
	pairstate => '',
	symcli_ver => '0',
	symcli_db_file => ''
);


sub new {
    my($proto, $resname, $name, $symhome, $mode, $devfotime, $iscompositegroup, $swaproles, $symcli_db_file) = @_;
    my $class = ref($proto) || $proto;
    my $self = { %default };
    my @RDFG;
    my $numofRDFGs;
    my (%arraydevs)  = ();
    my $func = "SRDFAgent::new";

    $self->{'resname'} = $resname;
    $self->{'name'} = $name;
    $self->{'home'} = $symhome;
    $self->{'mode'} = $mode;
    $self->{'iscompositegroup'} = $iscompositegroup;
    $self->{'swaproles'} = $swaproles;
    $self->{'start_time'} = $^T;
    if ((defined $symcli_db_file) && ($symcli_db_file !~ /^\s*$/)) {
	    SRDFAgent::set_db_file($symcli_db_file);
    }
    if ($self->{'mode'} eq ''){
	    $self->{'mode'} = "MODE_NOT_SPEC";
    }
    if ($self->{'iscompositegroup'} == 0){
	    $self->{'type'} = " -g ";
    } else {
	    $self->{'type'} = " -cg ";
    }
#swaprole = 0 ==> don't swap the role
#swaprole = 1 ==> check array capability before role swap.
#swaprole = 2 => don't check array capability. swap role directly assuming that administrator is aware 
    if($self->{'swaproles'}) { #swaprole attribute set to 1 or 2
	    $self->{'swapcapable'} = 1 ;
    }

    my $osType = $^O;
    if ( $osType =~ /^MSWin/i)
    {
        $self->{'home'} = "\"" . Win32::GetShortPathName($symhome) . "\"";
        $vcs_home = "\"" . Win32::GetShortPathName($VCSHOME) . "\"";
        $symrdf = "$self->{'home'}\\symrdf";
        $symdg = "$self->{'home'}\\symdg";
        $sympd = "$self->{'home'}\\sympd";
        $symcfg = "$self->{'home'}\\symcfg";
        $symcg = "$self->{'home'}\\symcg";
        $symdev = "$self->{'home'}\\symdev";
        $symcli = "$self->{'home'}\\symcli";
	$syminq = "$self->{'home'}\\syminq";
    }
    else
    {
        $self->{'home'} = $symhome;
        $symrdf = "$self->{'home'}/bin/symrdf";
        $symdg = "$self->{'home'}/bin/symdg";
        $sympd = "$self->{'home'}/bin/sympd";
        $symcfg = "$self->{'home'}/bin/symcfg";
        $symcg = "$self->{'home'}/bin/symcg";
        $symdev = "$self->{'home'}/bin/symdev";
        $symcli = "$self->{'home'}/bin/symcli";
	$syminq = "$self->{'home'}/bin/syminq";
    }

    $self->{'consistency'} = "Disabled";
    $self->{'devfotime'} = $devfotime;
    $self->{ARRAYDEVS} = \%arraydevs;

    $self->{'concurrent'} = -1;
    $self->{'rdfgopt'} = "";
    $self->{'symcli_ver'} = 0;
    bless($self, $class);
    return $self;
}

sub DESTROY {}

sub _exit
{
    my ($self, $stat) = @_;
    $self->unlock_res();
    exit($stat);
}

sub getAdvOpts {
    shift;
    my @args = @_;
    my ( $ret, $numAdOpts, $advOptsIndex, $loop, %advOpts );

    ( $ret, $numAdOpts, $advOptsIndex ) = VCSAG_GET_ATTR_VALUE ( 'AdvancedOpts', @args );
    
    for ( $loop = 1; $loop <= $numAdOpts; $loop++ ) {
        my ($key, $value);
        ($ret, $key) = VCSAG_GET_ATTR_VALUE ('AdvancedOpts', $advOptsIndex, $loop, @ARGV);
        $loop++;
        ($ret, $value) = VCSAG_GET_ATTR_VALUE ('AdvancedOpts', $advOptsIndex, $loop, @ARGV);
        $advOpts {$key} = $value;
    }
    return %advOpts;
}

sub Rep_Tech_RPO {
    my $ResName = shift;
    my ($groupname, $iscompositegroup, $symhome, $time, $ret, $time_behind);

    ($ret, $symhome) = VCSAG_GET_ATTR_VALUE ("SymHome", -1, 1, @ARGV);
    exit 0 if ($ret != $VCSAG_SUCCESS);
    ($ret, $groupname) = VCSAG_GET_ATTR_VALUE ("GrpName", -1, 1, @ARGV);
    exit 0 if ($ret != $VCSAG_SUCCESS);
    ($ret, $iscompositegroup) = VCSAG_GET_ATTR_VALUE ("IsCompositeGroup", -1, 1, @ARGV);
    exit 0 if ($ret != $VCSAG_SUCCESS);

    if ($iscompositegroup){
        $time_behind = `$symhome/bin/symcg show $groupname | grep 'Time that R2 is behind R1'`;
    } else {
    	$time_behind = `$symhome/bin/symdg show $groupname | grep 'Time that R2 is behind R1'`;
    }  	
    $time_behind =~ /\s+Time\ that\ (.*\w+)\s+:\s+(.*:.*:.*)\s+$/;
    $time = "Rep_Tech_RPO: "."$1"." by"." $2"." [hh:mm:ss] time as reported by Symmetrix.";
    return $time;
}

#Populates all the path information in the the devices array for a $dev. On linux powerpath commands are installed at /sbin/. 
#On other platform path may differ.
sub get_all_paths {
    my $func = 'get_all_paths';
    VCSAG_LOGDBG_MSG(1, "$func() Entered");
    my $dev = shift;
    my $devices = shift;
    my $powermt_cmd = "$powermt display dev=$dev";
    my @powermt_output;
    run_cmd ($powermt_cmd, \@powermt_output, 1, $func);
    foreach my $line (@powermt_output) {
        next if ($line =~ m /^$|^Pseudo|^Symmetrix|^Logical|^state|^=|^#|^-/);
        my @dev_array = split (/\s+/, $line);
        push (@$devices, $dev_array[3]);
    }
    VCSAG_LOGDBG_MSG(1, "$func() Exited");
}

# First check for type of multipathing.
# After multipathing is found get all the devices associated.
sub reload_devicemap {
    my $self = shift;
    my $func = "reload_devicemap";
    return if ($^O !~ /linux/i);
    VCSAG_LOGDBG_MSG(1, "$func() Entered");
	my $native_multipath_cmd = "$native_multipathing -ll";
    my $powermt_version = "$powermt version 1> $DEVNULL";
    my $multipath_output;
    my $updated=0;
	my @pdevs = $self->get_pdev_name();
	VCSAG_LOGDBG_MSG(1, "Checking for powerpath.");
	run_cmd ($powermt_version, 0, 1, $func);
	if ($? == 0) {
		VCSAG_LOG_MSG ("I", "Powerpath detected", 1112);
		foreach my $dev (@pdevs) {
			my @devices;
			get_all_paths($dev, \@devices);
			s/^/\/dev\// for @devices;
			push (@devices, "/dev/$dev");
			foreach my $dev_path (@devices) {
				my $cmd = "$blockdev --rereadpt $dev_path";
				run_cmd ($cmd, 0, 1, $func);
			}
		}
		return 0;
	} else {
		VCSAG_LOGDBG_MSG(1, "Rescanning devices [@pdevs].");
		foreach my $dev (@pdevs){
			my $rescan_cmd = "echo 1 > /sys/block/$dev/device/rescan";
			VCSAG_LOGDBG_MSG(1, "Issuing rescan command for [$dev]");
			`$rescan_cmd`;
		}
		VCSAG_LOGDBG_MSG(1, "Powerpath not found. Checking for native multipathing.");
		run_cmd($native_multipath_cmd, \$multipath_output, 1, $func);
		if ($multipath_output && $? == 0) {
			VCSAG_LOGDBG_MSG(1, "Native multipathing detected.");
			if ($#pdevs >= 0) {
				foreach my $dev (@pdevs){
					if (grep ($dev, $multipath_output)){
						VCSAG_LOGDBG_MSG (1, "Issuing command to reload device map");
        				run_cmd ("multipath -r", 0, 1, $func);
						last;
					}
				}
			}
		} else {
    	    VCSAG_LOGDBG_MSG (1, "No multipathing detected.");
    	}
	}
    VCSAG_LOGDBG_MSG(1, "$func() Exited");
	return 0;
}

sub get_pdev_name {
    my $self = shift;
    my $cmd;
    if ($self->{'iscompositegroup'} == 0) {
        $cmd = "$symdg show $self->{'name'} | grep '\/dev.*RDF'";
    } else { 
        $cmd = "$symcg show $self->{'name'} | grep '\/dev.*RDF'";
    }
    my (@dg_show_output, @pdevs);
    my $func = "get_pdev_name";
    VCSAG_LOGDBG_MSG(1, "$func() Entered");
    run_cmd($cmd, \@dg_show_output, 1, $func);
    foreach my $line (@dg_show_output) {
        my @dev_array = split (/\s+/, $line);
        my @device_path = split(/\//,$dev_array[2]);
        push (@pdevs, $device_path[-1]);
    }
    if ( $#pdevs == -1) {
		VCSAG_LOGDBG_MSG (1, "Checking if SYMAPI SERVER is configured");
        my ($gvdid, $gvdid_cmd) = ("$vcs_home/../VRTSsfmh/bin/gvdid", '');
        my %dev_to_vdid;
        my @gvdid_out;
        if ( $ENV{SYMCLI_CONNECT}) { 
			# SYMCLI_CONNECT is set during online. look get_symms subroutine
            my $ls_dev_cmd = "/bin/ls /sys/block | grep ^sd";
            my @all_dev = `$ls_dev_cmd`;
            foreach my $dev (@all_dev) {
                if ($dev !~ m/\d/) {
					chomp $dev;
                    $gvdid_cmd = "$gvdid -d /dev/$dev";
                    run_cmd ($gvdid_cmd, \@gvdid_out, 1, $func); 
					@gvdid_out = `$gvdid_cmd`;
                    foreach my $line (@gvdid_out) {
                       	if ($line =~ m/^(\/dev\/$dev)\|\+\+\|(EMC\w+)\|.*/) {
							$dev_to_vdid {$2} = "$dev";
                        } 
                    }
                }
            }
            my $computed_gvdid = '';
            for (my $index=0; $index < $self->{'numofRDFGs'}; $index++) {
                my $localKey = "Local" . "$self->{RDFG}{$index}" . "$self->{LSymmetrix}{$index}";
                foreach my $k (keys (%{$self->{ARRAYDEVS}})) {
                    if ($k eq $localKey) {
                        foreach my $j (keys (%{$self->{ARRAYDEVS}{$k}})) {
                        	$computed_gvdid = 'EMC' . "$self->{LSymmetrix}{$index}" . "$self->{ARRAYDEVS}{$k}{$j}";
                        	if (exists $dev_to_vdid{$computed_gvdid} ) {
                            	push (@pdevs, $dev_to_vdid{$computed_gvdid});
                        	}
						}
                    }
                }
            } 
        } else {
			VCSAG_LOGDBG_MSG (1, "Unable to determine the block devices. No rescanning possible.");
		}
    }
    VCSAG_LOGDBG_MSG(1, "Following pdevs are associated with the group [@pdevs].");
    VCSAG_LOGDBG_MSG(1, "$func() Exited");
    return @pdevs;
}

sub get_state {
    my $self = shift;
    my $ret;
    my $indexRDFG = 0;
    my $indexLSymmID = 0;
    my $indexRSymmID = 0;
    my $key;
    my $devidx = 0;
    my $output;
    my $func = "get_state";
    my $cmd = "$symrdf $self->{'type'} $self->{'name'} $self->{'rdfgopt'} query";
    run_cmd($cmd, \$output, 1, $func);
    my @out;
    my $modepat= qr/[SAD][X\.]/;
    if($self->{'symcli_ver'}<9.0) { 
	    $modepat= qr/[SAC]\./;
    }
    if(!$output) {
        $ret = "$?";
        my $cmd_txt=$cmd;
        $cmd_txt =~ s/\"//g;
        VCSAG_LOG_MSG("E", "Command $cmd_txt failed with error code:$ret\n", 1008, "$cmd_txt", "$ret");
        $self->_exit(0);
    }

    my @query = split('\n', $output);
    $self->{'state'} = "";
    foreach my $q (@query) {
        if ($q =~ /Group Number/) {
            @out = split('\s+', $q);
            $self->{RDFG}{$indexRDFG} = $out[5];
            $indexRDFG++;
            $devidx = 0;
            next;
        }
        if ($q =~ /^Remote Symmetrix ID/) {
            @out = split('\s+', $q);
            $self->{RSymmetrix}{$indexRSymmID} = $out[4];
            $indexRSymmID++;
            next;
        }
#
# a local symmetrix id can appear in these forms:
# DG's Symmetrix ID   : 000190300835    (Microcode  Version: 5773)
# OR
# Symmetrix ID        : 000190300835    (Microcode  Version: 5773)
#
        if ($q =~ /^Symmetrix ID/ or $q =~ /^DG's Symmetrix ID/) {
            my @out2 = split(':', $q);
            @out = split('\s+', $out2[1]);
            $self->{LSymmetrix}{$indexLSymmID} = $out[1];
            $indexLSymmID++;
            next;
        }
        if ($q =~ /Source/) {
            my @roleinfo = split('\s+', $q);
            if ($roleinfo[1] eq "Source") {
                $self->{'role'} = "r1";
            } else {
                $self->{'role'} = "r2";
            }
            next;
        }
        if (!($q =~/$modepat/)) {
            next;
        }

        my @stateinfo = split('\s+', $q);

        if ($stateinfo[10] =~ /S/) {
            $self->{'dcv_mode'} = "Synchronous";
        } elsif ($stateinfo[10] =~ /A/) {
            $self->{'dcv_mode'} = "Asynchronous"; 
        }

	if ($self->{'iscompositegroup'} == 1) {
		my @maceflags = split(//, $stateinfo[10]);
		VCSAG_LOGDBG_MSG(1, "Checking for consistency state");
		if($self->{'symcli_ver'}<9.0){
			if ($maceflags[2] =~ /X/) {
				VCSAG_LOGDBG_MSG(1, "Consistency is Enabled");
				$self->{'consistency'} = "Enabled";
			} elsif ($maceflags[2] =~ /./) {
				VCSAG_LOGDBG_MSG(1, "Consistency is Disabled");
				$self->{'consistency'} = "Disabled";
			}
		}
		else { 
			if ($maceflags[1] =~ /X/) {
				VCSAG_LOGDBG_MSG(1, "Consistency is Enabled");
				$self->{'consistency'} = 1;
			}
			elsif ($maceflags[1] =~ /./) {
				VCSAG_LOGDBG_MSG(1, "Consistency is Disabled");
				$self->{'consistency'} = 0;
			}
		}
	}
	{ # Adding the local and remote devices device id.  
		$key = "Local" . $self->{RDFG}{$indexRDFG - 1} .  $self->{LSymmetrix}{$indexLSymmID - 1};
		$self->{'ARRAYDEVS'}{$key}{$devidx} = $stateinfo[1];
		$key = "Remote" . $self->{RDFG}{$indexRDFG - 1} .  $self->{RSymmetrix}{$indexRSymmID - 1};
		$self->{'ARRAYDEVS'}{$key}{$devidx} = $stateinfo[6];
		$devidx++;
	}

        if ($stateinfo[2] eq "RW") {
            if ($self->{'state'} eq "WD") {
                $self->{'state'} = "Invalid";
                return;
            } else {
                $self->{'state'} = "RW";
            }
        }
        if ($stateinfo[2] eq "WD") {
            if ($self->{'state'} eq "RW") {
                $self->{'state'} = "Invalid";
                return;
            } else {
                $self->{'state'} = "WD";
            }
        }

        ++$self->{'devs'};
        $self->{'invalid'} = $self->{'invalid'} + $stateinfo[3];
    }
    if (!$self->{'state'}) {
        $self->{'state'} = "Invalid";
    }
    $self->{'numofRDFGs'} = $indexRDFG;
}

sub get_mode {
    my $self = shift;
    my $retVal = 0; # 0 stands for Synchronous
    my $dcv_mode = $self->{'dcv_mode'};
    my $func = "get_mode";

	# Check the specified mode with the discovered mode. The two should match. 
	# If not, then give an error and exit.
	#
	# Partitioned state is a very special case. 
	#
	# In partitioned state, for asynchronous mode of replication, the mode on
	# the R2 side changes to Synchronous (actually it should drop to adaptive
	# copy, but it drops to the default of "S" i.e. synchronous.)
	#
	# To consider the above case, we do not signal an error if -
	#
	# - the devices are in a partitioned state.
	# - The mode specified in the configuration is 'Asynchronous'
	# - The discovered mode is 'Synchronous'
	#
	# However, in a partitioned state, we will signal an error if -
	#
	# 1. The configured state is 'Synchronous' and the discovered state 
	#    is 'Asynchronous'
	#
	# +-------------------------------------------------------------+
	# |   Configured State  |  Discovered State | Result            |
	# |---------------------+-------------------+-------------------|
	# |                     |                   |                   |
	# |   Synchronous       | Synchronous       | Valid             |
	# |---------------------+-------------------+-------------------|
	# |   Synchronous       | Asynchronous      | Error             |
	# |---------------------+-------------------+-------------------|
	# |   Asynchronous      | Synchronous       | Valid if          |
	# |                     |                   | partitioned state |
	# |                     |                   |                   |
	# |                     |                   | Error if not      |
	# |                     |                   | partitioned state |
	# |                     |                   |                   |
	# |---------------------+-------------------+-------------------|
	# |   Asynchronous      | Asynchronous      | Valid             |
	# +-------------------------------------------------------------+
	#
	if ($self->{'mode'} ne "MODE_NOT_SPEC") {

        if (($self->{'mode'} eq $dcv_mode) && ($self->{'mode'} eq "Asynchronous")) {
            $retVal = 1;
        }

        if (($self->{'mode'} ne $dcv_mode) && ($self->{'mode'} eq "Synchronous")) {
# Error - Configured is Sync and discovered is Async.
                VCSAG_LOG_MSG("E", "SRDF replication mode conflict detected. Mode of replication configured is $self->{'mode'}, while the device group is in $dcv_mode mode. Cannot proceed with online of $self->{'resname'} resource. Administrative intervention is required.", 91, "$self->{'mode'}", "$dcv_mode", "$self->{'resname'}");
            $self->_exit(0);
        }
        if (($self->{'mode'} ne $dcv_mode) && ($self->{'mode'} eq "Asynchronous")) {
# Configured is Asynch and discovered is sync.
            my $partitioned = run_cmd("$symrdf $self->{'type'} $self->{'name'} $self->{'rdfgopt'} -partitioned verify > $DEVNULL 2>&1", 0, 1, $func);
            if ($partitioned != 0) {
# Error if not partitioned.
                VCSAG_LOG_MSG("E", "SRDF replication mode conflict detected. Mode of replication configured is $self->{'mode'}, while the device group is in $dcv_mode mode. Cannot proceed with online of $self->{'resname'} resource. Administrative intervention is required.", 89, "$self->{'mode'}", "$dcv_mode", "$self->{'resname'}");
                $self->_exit(0);
            }
            VCSAG_LOG_MSG("N", "Devices in group $self->{'grpname'} are in partitioned state, mode of replication configured is $self->{'mode'}, detected is $dcv_mode", 26, "$self->{'name'}", "$self->{'mode'}", "$dcv_mode");
            $retVal = 1;
        }
    } else {
        if ($dcv_mode eq "Asynchronous") {
            $retVal = 1;
        }
    }
    return $retVal;
}

sub get_consistency {
    my $self = shift;
    my $ret = 0;
    my $func = "get_consistency";
    my @query;
    if ($self->{'iscompositegroup'} == 0){
        run_cmd("$symdg show $self->{'name'}", \@query, 1, $func);
        foreach my $q (@query) {
            if (($q =~ /Session Consistency State/) && ($q =~ /Enabled/)) {
                return 1;
            }
        }
    }
    return 0;
}

sub failover_required {
    my ($self) = shift;
    my $func = "failover_required";
    if ($self->{'state'} eq "Invalid") {
        return 2;
    }

    if (($self->{'state'} eq "RW") && ($self->{'role'} eq "r2")) {
        my $ret = run_cmd("$symrdf $self->{'type'} $self->{'name'} $self->{'rdfgopt'} -split verify >$DEVNULL 2>&1", 0, 1, $func);
        if ($ret == 0) {
            return 3;
        }
    }
    if ($self->{'state'} eq "WD") {
        my $ret = run_cmd("$symrdf $self->{'type'} $self->{'name'} $self->{'rdfgopt'} -suspended verify >$DEVNULL 2>&1", 0, 1, $func);
        if ($ret == 0) {
            return 4;
        }
        return 1;
    }
    return 0;
}

sub failover_time {
    my ($self) = @_;
    my $cur_time = time();
    my $remaining_time = ($self->{'onlinetimeout'} - ($cur_time - $self->{'start_time'}));
    my $failover_time = $self->{'devs'} * $self->{'devfotime'};
    my $count = int(($remaining_time - ($failover_time + 10)) / 5);
    if ($count) {
        $count = $count - 1;
    }
    return $count;
}

sub get_count {
    my ($self) = @_;
    my $cur_time = time();
    my $func = "get_count";
	#
	# Get the resource-specific OnlineTimeout if set, else get the SRDF
	# type specific value
	#
    if ($self->{'onlinetimeout'} == 0) {
        my @onlinetimeout;
        if (run_cmd("$ag_i18n_inc::HARES -value $self->{'resname'} OnlineTimeout", 
                    \@onlinetimeout, 1, $func)) {
            run_cmd("$ag_i18n_inc::HATYPE -value SRDF OnlineTimeout", 
                    \@onlinetimeout, $func);
        }
        $self->{'onlinetimeout'} = $onlinetimeout[0];
    }
    my $remaining_time = ($self->{'onlinetimeout'} - ($cur_time - $self->{'start_time'}));
    my $count = int($remaining_time / 5) - 1;
    if ($count <= 0) {
        VCSAG_LOG_MSG("W", "not enough time to issue the symcli commands; administrative intervention or automatic restart required", 60);
        $self->_exit(0);
    }
    return $count;
}

sub wait_for_sync {
    my ($self) = @_;
    my $count = $self->get_count();
    my $halfCount;
    my $func = "wait_for_sync";
    if ($self->{'role'} eq "r1") {
        my $ret = run_cmd("$symrdf $self->{'type'} $self->{'name'} $self->{'rdfgopt'} -updateinprog verify >$DEVNULL 2>&1", 0, 1, $func);
        $ret = $ret / 256;
        if ($ret != 36) {
            VCSAG_LOG_MSG("N", "failback of group $self->{'name'} will wait until all updates in progress complete or entry point times out", 75, "$self->{'name'}");
            my $ret = run_cmd("$symrdf $self->{'type'} $self->{'name'} $self->{'rdfgopt'} -c $count -i 5 -updated verify >$DEVNULL 2>&1", 0, 1, $func);
            if ($ret) {
                return 1;
            }
        }
    } else {
        my $ret = run_cmd("$symrdf $self->{'type'} $self->{'name'} $self->{'rdfgopt'} -syncinprog verify >$DEVNULL 2>&1", 0, 1, $func);
        $ret = $ret / 256;
        if ($ret != 28) {
            VCSAG_LOG_MSG("N", "failover of group $self->{'name'} will wait until all synchronization in progress complete or entry point times out", 62, "$self->{'name'}");
            if ($self->get_mode()){
                $halfCount = int($count / 2);
                $ret = run_cmd("$symrdf $self->{'type'} $self->{'name'} $self->{'rdfgopt'} -c $halfCount -i 5 -consistent verify >$DEVNULL 2>&1", 0, 1, $func);
                if ($ret) {
                    $ret = run_cmd("$symrdf $self->{'type'} $self->{'name'} $self->{'rdfgopt'} -c $halfCount -i 5 -synchronized verify >$DEVNULL 2>&1", 0, 1, $func);
                }
            } else {
                $ret = run_cmd("$symrdf $self->{'type'} $self->{'name'} $self->{'rdfgopt'} -c $count -i 5 -synchronized verify >$DEVNULL 2>&1", 0, 1, $func);
            }

            if ($ret) {
                return 1;
            }
        }
    }
    return 0;
}

# Delete the vmdisks corresponding to the disks in the SRDF group.
sub delete_vm_disks
{
    my ($self) = @_;
    my $func = "delete_vm_disks";
    if ($OS ne "AIX") {
        return;
    }
# We cannot get the symdev - OS Handle mapping from remote symapi server.
# Hence return from here if SYMCLI_CONNECT is set.
    if ($ENV{SYMCLI_CONNECT}) {
        VCSAG_LOG_MSG("I", "Skipping processing of vmdisks on AIX since SYMCLI_CONNECT is set.", 50);
        return;
    }

 if(!$self->check_dmp_pkg_level()) {
        VCSAG_LOG_MSG("I", "Skip processing of vxdisk rm as expected vxvm pkg level is present.", 50);
	return;
 }

    VCSAG_LOGDBG_MSG(1, "$func() Entered");
    my @local_symms = ();
    my %sym_pdevs = ();
    for my $symm (keys %{$self->{LSymmetrix}}) {
        my $sid = $self->{LSymmetrix}{$symm};
        push(@local_symms, $sid);

# Collect 'sympd list' o/p for each array.
        my @sympd_lines;
        run_cmd("$self->{'home'}/bin/sympd list -sid $sid |awk '/^\\/dev\\/[a-zA-Z0-9_]* /{print \$2,\$1}'",
                 \@sympd_lines, 1, $func); #Eliminating the multipathed devices. Capturing only the devices in /dev/ directory to avoid dupliactes entries of the same device
        s/rhdisk/hdisk/g foreach (@sympd_lines); # Replace rhdisk with hdisk since vxdisk list contains hdisk only.
            $sym_pdevs{$sid} = [ @sympd_lines ];
    }
# Find symm devices in each Symmetrix participating in the RDF group.
    my %grpdevs_hash = ();
    if(! $self->_get_grp_devs($self->{'type'}, \%grpdevs_hash, @local_symms)) {
        VCSAG_LOG_MSG("E", "delete_vm_disks: Could not find devices specific to RDF Group.", 86);
        return;
    }
# Find the pdev names of the devices participating in the RDF group.
    my @all_paths = ();
    for my $sid (keys %grpdevs_hash) {
        my $dev_list_pat = join('|', @{$grpdevs_hash{$sid}});
        my @pdevs = grep (/$dev_list_pat/, @{$sym_pdevs{$sid}});
        s/\S+\s+.*(hdisk.*)/^$1 / foreach (@pdevs); # Extract the pdev name from the path and create a pattern for matching in the vxdisk list o/p.;
        my @pdevs_raw = grep (/hdisk/, @pdevs);
        push(@all_paths, @pdevs_raw);
    }

    my @vxdisks;
    my ($awk_buff, $itr) = (399, 0);
    my $awk_regex = $all_paths[0] if ($#all_paths != -1);
    $itr++;

    # awk fails if more then 399 bytes are present in the regex on AIX. To avoid the failure processing in batches
    while ($itr < ($#all_paths + 1)) {
        my $disk_len = length ($all_paths[$itr]);
        if (length($awk_regex) + 2 + $disk_len <= $awk_buff){
            $awk_regex .= " |$all_paths[$itr]";
            $itr++;
            if ($itr >= ($#all_paths +1)){
                my @temp_arr;
                run_cmd("$VXDISK path |awk '/$awk_regex/ {print \$2}' |sort -u", \@temp_arr, 1, $func);
                push (@vxdisks, @temp_arr);
            }
        } else {
            my @temp_arr;
            run_cmd("$VXDISK path |awk '/$awk_regex/ {print \$2}' |sort -u", \@temp_arr, 1, $func);
            push (@vxdisks, @temp_arr);
            $awk_regex = $all_paths[$itr];
            $itr++;
        }
    }
    my ($buf_len, $i, $op) = (256, 0);
    my $cmd = "$VXDISK rm ";
    my $cmd_len = length($cmd);
    my $num_disks = $#vxdisks + 1;
    while ($i < $num_disks) {
        my $dn_l = length($vxdisks[$i]) + 1;
        if (($cmd_len + $dn_l) <= $buf_len) {
            $cmd .= "$vxdisks[$i] ";
            $cmd_len += $dn_l;
            $i++;
            if ($i >= $num_disks) {
                my $ret = run_cmd($cmd, \$op, 1, $func);
                if ($ret) {
                    VCSAG_LOG_MSG("W", "vxdisk removal command [$cmd] failed with error code: [$ret], o/p: [$op]", 88, "$cmd", "$ret", "$op");
                }
                last;
            }
        } else {
            my $ret = run_cmd($cmd, \$op, 1, $func);
            if ($ret) {
                VCSAG_LOG_MSG("W", "vxdisk removal command [$cmd] failed with error code: [$ret], o/p: [$op]", 88, "$cmd", "$ret", "$op");
            }
            $cmd = "$VXDISK rm ";
            $cmd_len = length($cmd);
        }
    }
    VCSAG_LOGDBG_MSG(1, "$func() Exiting.");
}

# This function deletes the vmdisks corresponding to the luns in the device
# group and then rescanning the vxvm subsystem to recreate the lost vmdsks.
# This is done only on AIX since on AIX, vxvm caches the "readonly"ness of
# the devices even after failing over.
# For Solaries and RHL only do scandisks changes and return after that.  
sub vxvm_scan
{
    my ($self) = @_;
    my $func = "vxvm_scan";
    if ($OS eq "Windows NT") {
        return;
    }
    
    my ($sectivescan, $ret);
    $ret = $self->get_devices(\$sectivescan);
    if ($OS eq "AIX") {
	VCSAG_LOGDBG_MSG(3, "$func: calling delete_vm_disks");
	$self->delete_vm_disks();
	}
    if($ret == 0){
         VCSAG_LOG_MSG("I", "cli[$sectivescan]", 1126, "$sectivescan");
	 $ret = run_cmd("$sectivescan", 0, 1, $func);
	if($ret != 0){
	 #vxdisk scandisk device= CLI failed hence call it for all devices
         VCSAG_LOG_MSG("I", "Selective devices cli failed going for : $VXDISK scandisks ", 1125, "$VXDISK");
	 run_cmd("$VXDISK scandisks", 0, 1, $func);
	}

    }
    else{
	 #failed to get selective devices hence calling vxdisk scandisks cli
         VCSAG_LOG_MSG("I", "Could not get selective devices. going for : $VXDISK scandisks ", 1122, "$VXDISK");
	 run_cmd("$VXDISK scandisks", 0, 1, $func);
	}

    if ($OS ne "AIX") {
	return ;
	}
    my ($grp) = split('\n', `$ag_i18n_inc::HARES -value $self->{'resname'} Group`);
    my ($para) = split('\n', `$ag_i18n_inc::HAGRP -value $grp Parallel`);
    return 0 if ($para != 1);

# Find all nodes in the SystemList and issue vxvm_rescan on them
    my (@locsysarr);
    run_cmd("$ag_i18n_inc::HASYS -nodeid", \@locsysarr, 1, $func);
    my @tokens = split(/\s+/, $locsysarr[0]);
    my ($thisnode) = $tokens[5];

    my ($grp, $syslist_str);
    run_cmd("$ag_i18n_inc::HARES -value $self->{'resname'} Group", \$grp, 1, $func);
    run_cmd("$ag_i18n_inc::HAGRP -value $grp SystemList", \$syslist_str, 1, $func);
    my @syslist = split(/\s+/, $syslist_str);
    foreach my $sys (@syslist) {
        next if ($sys =~ /^\d+$/);
        next if ($sys eq $thisnode);
        run_cmd("$ag_i18n_inc::HARES -action $self->{'resname'} vxvm_scan -sys $sys", 0, 1, $func);
    }
}

sub check_invalid_r21
{
    my ($self, $op) = @_;

    my $ret = 0;
    if ($op =~ /The other pair of the R\d\d is not in a valid RDF state for this operation./) {
        self->validate_remote_r21();
# TODO: Check return status.
        $ret = 1;
    }
    return $ret;
}

sub run_cmd_chk_err
{
    my ($self, $cmd, $func) = @_;

    my $ret = 0;
    my $op;
    $ret = run_cmd($cmd, \$op, 1, $func);
    if($ret) {
        if ($self->check_invalid_r21($op)) {
# Reissue the failover command again.
            $ret = system($cmd);
            if ($ret) {
                $ret = $ret / 256;
            }
        } else {
            $ret = $? / 256;
        }
    }
    return $ret;
}

sub takeover {
    my ($self, $autotakeover) = @_;
    my $count;
    my $failovercount;
    my $func = "takeover";

	#
	# Get the resource-specific OnlineTimeout if set, else get the SRDF
	# type specific value
	#
    if ($self->{'onlinetimeout'} == 0) {
        my @onlinetimeout;
        if (run_cmd("$ag_i18n_inc::HARES -value $self->{'resname'} OnlineTimeout", 
                    \@onlinetimeout, 1, $func)) {
            run_cmd("$ag_i18n_inc::HATYPE -value SRDF OnlineTimeout", 
                    \@onlinetimeout, $func);
        }
        $self->{'onlinetimeout'} = $onlinetimeout[0];
    }

    if ($OS ne "Windows NT") {
        my $ret = run_cmd("/usr/bin/vmware -v", 0, 1, $func);
# If command succeeds, we will assume that it is vmware system
        if ($ret == 0) {
            $self->{'is_vmware'} = 1;
        }
    }

    my $ret = run_cmd("$symrdf $self->{'type'} $self->{'name'} $self->{'rdfgopt'} -partitioned verify > $DEVNULL 2>&1", 0, 1, $func);
    if ($ret == 0) {
        if ($autotakeover) {
# Synchronous mode and invalid tracks found.
            if (!($self->get_mode()) && ($self->{'invalid'})) {
                VCSAG_LOG_MSG("E", "devices in group $self->{'name'} are in Partitioned state and invalid tracks exist; takeover cannot occur and administrative intervention to restore consistent data is required", 11, "$self->{'name'}");
                $self->_exit(0);
            } else {
# Asynchronous mode and invalid tracks found.
# OR
# Synchronous mode and no invalid tracks found.
                $failovercount = $self->failover_time();
                if ($failovercount <= 0) {
                    VCSAG_LOG_MSG("W", "not enough time to issue failover command; administrative intervention or automatic restart required", 27);
                    $self->_exit(0);
                }

                VCSAG_LOG_MSG("N", "devices in group $self->{'name'} are in Partitioned state, and autotakeover is enabled; failover of devices will proceed", 41, "$self->{'name'}");
                my $ret = $self->run_cmd_chk_err("$symrdf $self->{'type'} $self->{'name'} $self->{'rdfgopt'} -noprompt -c $failovercount -i 5 failover 2>&1",
                        $func);
                if($ret) {
                    VCSAG_LOG_MSG("E", "failover of Partitioned devices in group $self->{'name'} failed; administrative intervention is required", 12, "$self->{'name'}");
                    $self->_exit(0);
                }
                VCSAG_LOG_MSG("N", "failover of devices in group $self->{'name'} succeeded", 23, "$self->{'name'}");
                $self->vxvm_scan();
                $self->create_lockfile();
                if ($self->{'is_vmware'}) {
                    $self->vmware_online();
                } else { 
                    $self->cvm_import();
                }
                $self->reload_devicemap();
                $self->_exit(0);
            }
        } else {
            VCSAG_LOG_MSG("E", "devices in group $self->{'name'} are in Partitioned state but autotakeover is not enabled; administrative intervention is required", 13, "$self->{'name'}");
            $self->_exit(0);
        }
    }
}

sub failback {
    my ($self) = @_;
    my $count;
    my $halfCount;
    my $failbackcount;
    my $func = "failback";

	#
	# Get the resource-specific OnlineTimeout if set, else get the SRDF
	# type specific value
	#
    if ($self->{'onlinetimeout'} == 0) {
        my @onlinetimeout;
        if (run_cmd("$ag_i18n_inc::HARES -value $self->{'resname'} OnlineTimeout", 
                    \@onlinetimeout, 1, $func)) {
            run_cmd("$ag_i18n_inc::HATYPE -value SRDF OnlineTimeout", 
                    \@onlinetimeout, $func);
        }
        $self->{'onlinetimeout'} = $onlinetimeout[0];
    }

    if ($OS ne "Windows NT") {
        my $ret = run_cmd("/usr/bin/vmware -v", 0, 1, $func);
# If command succeeds, we will assume that it is vmware system
        if ($ret == 0) {
            $self->{'is_vmware'} = 1;
        }
    }

    $count = $self->get_count();
    $halfCount = int($count / 2);

    my $ret;
    my $temp_counter = 1;               # imp to exit on temp_counter = half counter

    if($halfCount == 0) {
#Try to execute failedover and updated verification atleast once
	    $halfCount = 1;
    }
    while ($temp_counter <= $halfCount) {
	    VCSAG_LOGDBG_MSG(3, "Counter no: $temp_counter for failedover and updated verification");
	    my $ret = run_cmd("$symrdf $self->{'type'} $self->{'name'} $self->{'rdfgopt'} -failedover verify > $DEVNULL 2>&1", 0, 1, $func);
	    if ($ret) {
		    $count = $self->get_count();
		    $ret = run_cmd("$symrdf $self->{'type'} $self->{'name'} $self->{'rdfgopt'} -updated verify > $DEVNULL 2>&1", 0, 1, $func);
		    if ($ret) {
			    if($temp_counter == $halfCount) {
				    VCSAG_LOG_MSG("E", "all devices in group $self->{'name'} are not in Failed Over or Updated state in preparation for failback; administrative intervention is required", 14, "$self->{'name'}");
				    $self->_exit(0);
			    }
			    sleep (5);
			    $temp_counter = $temp_counter + 1;
			    next;
		    }
	    }
	    last;
    }

    $failbackcount = $self->failover_time();
    if ($failbackcount <= 0) {
        VCSAG_LOG_MSG("W", "not enough time to issue failback command; administrative intervention or automatic restart required", 28);
        $self->_exit(0);
    }

    VCSAG_LOG_MSG("N", "issuing failback of group $self->{'name'}", 63, "$self->{'name'}");
    $ret = run_cmd("$symrdf $self->{'type'} $self->{'name'} $self->{'rdfgopt'} -noprompt -i 5 -c $failbackcount failback", 0, 1, $func);

    if ($ret) {
        $ret = $ret / 256;
        VCSAG_LOG_MSG("E", "failback of group $self->{'name'} failed; error code is $ret", 16, "$self->{'name'}", "$ret");
        $self->_exit(0);
    }

    $self->get_state();
    if ($self->{'state'} eq "RW") {
        VCSAG_LOG_MSG("N", "failback of devices in group $self->{'name'} succeeded", 22, "$self->{'name'}");
        $self->vxvm_scan();
        $self->create_lockfile();
        if ($self->{'is_vmware'}) {
            $self->vmware_online();
        } else { 
            $self->cvm_import();
        }
        $self->reload_devicemap();
        $self->_exit(0);
    } else {
        VCSAG_LOG_MSG("E", "failback of group $self->{'name'} failed", 16, "$self->{'name'}");
        $self->_exit(0);
    }
}

sub validate_remote_r21 {
    my ($self) = @_;
    my $func = "validate_remote_r21";

    my $ret = run_cmd("$symrdf $self->{'type'} $self->{'name'} rw_enable r1 -hop2 -noprompt", 0, 1, $func);
    if ($ret) {
        VCSAG_LOG_MSG("E", "Cannot failover the SRDF replication. The SRDF agent could not change the status of the other pair of the R21 device in device group [$self->{'name'}]. Administrative intervention required.", 82, "$self->{'name'}");
        $self->_exit(0);
    }
}

sub failover {
    my ($self) = @_;
    my $count;
    my $halfCount;
    my $failovercount;

    my $ret = 0;
    my $disabled_consistency;
    my $index;
    my $fileName;
    my $func = "failover";

	#
	# Get the resource-specific OnlineTimeout if set, else get the SRDF
	# type specific value
	#
    if ($self->{'onlinetimeout'} == 0) {
        my @onlinetimeout;
        if (run_cmd("$ag_i18n_inc::HARES -value $self->{'resname'} OnlineTimeout", 
                    \@onlinetimeout, 1, $func)) {
            run_cmd("$ag_i18n_inc::HATYPE -value SRDF OnlineTimeout", 
                    \@onlinetimeout, $func);
        }
        $self->{'onlinetimeout'} = $onlinetimeout[0];
    }

    if ($OS ne "Windows NT") {
        my $ret = run_cmd("/usr/bin/vmware -v", 0, 1, $func);
# If command succeeds, we will assume that it is vmware system
        if ($ret == 0) {
            $self->{'is_vmware'} = 1;
        }
    }

    $count = $self->get_count();
    if ($self->get_mode()) {
        $halfCount = int($count / 2);
        $ret = run_cmd("$symrdf $self->{'type'} $self->{'name'} $self->{'rdfgopt'} -c $halfCount -i 5 -consistent verify > $DEVNULL 2>&1", 0, 1, $func);
        if ($ret) {
            $halfCount = int($count / 2);
            $count = $self->get_count();
            $ret = run_cmd("$symrdf $self->{'type'} $self->{'name'} $self->{'rdfgopt'} -c $halfCount -i 5 -synchronized verify >$DEVNULL 2>&1", 0, 1, $func);
        }
    } else {
        $ret = run_cmd("$symrdf $self->{'type'} $self->{'name'} $self->{'rdfgopt'} -c $count -i 5 -synchronized verify >$DEVNULL 2>&1", 0, 1, $func);
    }
    if ($ret) {
        VCSAG_LOG_MSG("W", "all devices in group $self->{'name'} are not in synchronized or consistent state in preparation for failover; administrative intervention is required", 17, "$self->{'name'}");
        $self->_exit(0);
    }

    $failovercount = $self->failover_time();
    if ($failovercount <= 0) {
        VCSAG_LOG_MSG("W", "not enough time to issue failover command; administrative intervention or automatic restart required", 27);
        $self->_exit(0);
    }

    if ($self->{'iscompositegroup'} == 1) {
        $count = $self->get_count();

        if ($self->{'consistency'} eq "Enabled") {
            VCSAG_LOG_MSG("N", "disabling consistency for group $self->{'name'}", 76, "$self->{'name'}");
            $ret = run_cmd("$symcg $self->{'type'} $self->{'name'} -i 5 -c $count -noprompt -force disable", 0, 1, $func);
            if ($ret) {
                VCSAG_LOG_MSG("E", "Disabling the consistency for the group $self->{'name'} failed; error code is $ret", 77, "$self->{'name'}", "$ret");
                $self->_exit(0);
            }
        } 
		

        VCSAG_LOG_MSG("N", "issuing failover of group $self->{'name'}", 18, "$self->{'name'}");
        my $ret = $self->run_cmd_chk_err("$symrdf $self->{'type'} $self->{'name'} $self->{'rdfgopt'} -noprompt -i 5 -c $failovercount failover 2>&1", $func);
        if($ret) {
            $ret = $ret / 256;
            VCSAG_LOG_MSG("E", "Failover of the group $self->{'name'} failed; error code is $ret", 78, "$self->{'name'}", "$ret");
            $self->_exit(0);
        }

        VCSAG_LOG_MSG("N", "suspending the links for group $self->{'name'}", 61, "$self->{'name'}");
        my @out;
        $ret = run_cmd("$symrdf $self->{'type'} $self->{'name'} $self->{'rdfgopt'} -i 5 -c $count -noprompt suspend 2>&1", \@out, 1, $func);
        $ret = $ret>>8;
		#
		# hack to overcome bug in SE7.0.X
		# The bug is in the return code of "symrdf ...suspend". It returns a 1
		# if already in the desired state insted of CLI_C_ALREADY_IN_STATE(18)
		#
        {
            $ret = $self->__hack_se7_0_X_bug1($ret, @out);
        }

        if($ret && $ret != 18) {
            VCSAG_LOG_MSG("E", "Suspension of the links for the group $self->{'name'} failed; error code is $ret", 55, "$self->{'name'}", "$ret");
            $self->_exit(0);
        }

        if ($self->{'swapcapable'}) {
            VCSAG_LOG_MSG("N", "swapping the roles of the devices for group $self->{'name'}", 64, "$self->{'name'}");
            for ($index=0; $index<$self->{'numofRDFGs'}; $index++) {
                $ret = $self->create_device_files(\$fileName, $self->{LSymmetrix}{$index}, $self->{RSymmetrix}{$index}, $self->{RDFG}{$index});
                if ($ret) {
                    VCSAG_LOG_MSG("E", "Error creating the device file for group $self->{'name'} ", 65, "$self->{'name'}");
                    $self->_exit(0); 
                }
                sleep 3;
                $count = $self->get_count();
                my $ret = $self->run_cmd_chk_err("$symrdf -f \"$fileName\" -rdfg $self->{RDFG}{$index} -sid $self->{LSymmetrix}{$index} swap -noprompt -i 5 -c $count", $func);
                if($ret) {
                    VCSAG_LOG_MSG("E", "Swapping the roles of the devices for the group $self->{'name'} failed; error code is $ret", 66, "$self->{'name'}", "$ret");
                    $ret = $self->remove_device_files($fileName);
                    if ($ret) {
                        VCSAG_LOG_MSG("E", "Error removing the device file for group $self->{'name'} ", 67, "$self->{'name'}");
                    }
                    $self->_exit(0);
                }
                $ret = $self->remove_device_files($fileName);
                if ($ret) {
                    VCSAG_LOG_MSG("E", "Error removing the device file for group $self->{'name'} ", 67, "$self->{'name'}");
                    $self->_exit(0);
                }
            }

            $count = $self->get_count();
            VCSAG_LOG_MSG("N", "issuing establish/resume of group $self->{'name'}", 69, "$self->{'name'}");
            my $ret = $self->run_cmd_chk_err("$symrdf $self->{'type'} $self->{'name'} $self->{'rdfgopt'} -noprompt -i 5 -c $count establish", $func);
            if($ret) {
                VCSAG_LOG_MSG("E", "Establish=link for the group $self->{'name'} failed; error code is $ret", 70, "$self->{'name'}", "$ret");
                $self->_exit(0);
            }
        }


	if ($self->{'consistency'} eq "Enabled") {
		$count = $self->get_count();
		VCSAG_LOG_MSG("N", "re-enable consistency for group $self->{'name'}", 71, "$self->{'name'}");
		$ret = system("$symcg $self->{'type'} $self->{'name'} -noprompt -i 5 -c $count enable");
		if ($ret) {
			VCSAG_LOG_MSG("N", "re-enabling consistency on group $self->{'name'} failed. Administrative intervention required.", 72, "$self->{'name'}");
			$self->_exit(0);
		}
	}
        $ret = system("$symcfg sync");
        if ($ret) {
            $ret = $ret / 256;
            VCSAG_LOG_MSG("E", "Synching of database failed; error code is $ret", 81, "$ret");
            $self->_exit(0);
        }
        sleep(3);
        $self->get_state();
        if ($self->{'state'} eq "RW") {
            VCSAG_LOG_MSG("N", "failover of devices in group $self->{'name'} succeeded", 23, "$self->{'name'}");
            $self->vxvm_scan();
            $self->create_lockfile();
            if ($self->{'is_vmware'}) {
                $self->vmware_online();
            } else { 
                $self->cvm_import();
            }
            $self->reload_devicemap();
            $self->_exit(0);
        } else {
            VCSAG_LOG_MSG("E", "failover of group $self->{'name'} failed , devices are not read-write enabled", 80, "$self->{'name'}");
            $self->_exit(0);
        }
    } else {
        if ($self->{'swapcapable'} == 0) {
# No Swap personality and normal failover.
            VCSAG_LOG_MSG("N", "issuing failover of group $self->{'name'}", 18, "$self->{'name'}");
            my $ret = $self->run_cmd_chk_err("$symrdf $self->{'type'} $self->{'name'} $self->{'rdfgopt'} -noprompt -i 5 -c $failovercount failover 2>&1", $func);
            if($ret) {
                VCSAG_LOG_MSG("E", "failover of device group $self->{'name'} failed; error code is $ret", 42, "$self->{'name'}", "$ret");
                $self->_exit(0);
            }
            sleep(3);
            $self->get_state();
            if ($self->{'state'} eq "RW") {
                VCSAG_LOG_MSG("N", "failover of devices in group $self->{'name'} succeeded", 23, "$self->{'name'}");
                $self->vxvm_scan();
                $self->create_lockfile();
                if ($self->{'is_vmware'}) {
                    $self->vmware_online();
                } else { 
                    $self->cvm_import();
                }
                $self->reload_devicemap();
                $self->_exit(0);
            } else {
                VCSAG_LOG_MSG("E", "failover of group $self->{'name'} failed, devices are not read-write enabled", 80, "$self->{'name'}");
                $self->_exit(0);
            }
        }

# For swap devices, if mode is sync.
        if ( !$self->get_mode ()) {
            VCSAG_LOG_MSG("N", "issuing failover/role swap of device group $self->{'name'}", 43, "$self->{'name'}");
            my $ret = $self->run_cmd_chk_err("$symrdf $self->{'type'} $self->{'name'} $self->{'rdfgopt'} -noprompt -i 5 -c $failovercount failover -establish 2>&1", $func);
            if($ret) {
                VCSAG_LOG_MSG("E", "failover of device group $self->{'name'} failed; error code is $ret", 42, "$self->{'name'}", "$ret");
                $self->_exit(0);
            }
            sleep(3);
            $self->get_state();
        } else {
            VCSAG_LOG_MSG("N", "issuing -force failover of device group $self->{'name'}", 44, "$self->{'name'}");
            my $ret = $self->run_cmd_chk_err("$symrdf $self->{'type'} $self->{'name'} $self->{'rdfgopt'} -noprompt -force -i 5 -c $failovercount failover 2>&1", $func);
            if($ret) {
                VCSAG_LOG_MSG("E", "failover of device group $self->{'name'} failed; error code is $ret", 42, "$self->{'name'}", "$ret");
                $self->_exit(0);
            }

            $count = $self->get_count();
            $disabled_consistency = 0;
            if ($self->get_consistency()) {
                $disabled_consistency = 1;
                VCSAG_LOG_MSG("N", "disabling consistency for device group $self->{'name'}", 45, "$self->{'name'}");
                $ret = run_cmd("$symrdf $self->{'type'} $self->{'name'} $self->{'rdfgopt'} -i 5 -c $count -noprompt disable >$DEVNULL", 0, 1, $func);
                if ($ret) {
                    VCSAG_LOG_MSG("E", "Disabling the consistency for the device group $self->{'name'} failed; error code is $ret", 52, "$self->{'name'}", "$ret");
                    $self->_exit(0);
                }
            }
            sleep(3);
            $count = $self->get_count();
            VCSAG_LOG_MSG("N", "issuing role swap of device group $self->{'name'}", 46, "$self->{'name'}");
            $ret = $self->run_cmd_chk_err("$symrdf $self->{'type'} $self->{'name'} $self->{'rdfgopt'} -noprompt -i 5 -c $count swap >$DEVNULL", $func);
            if($ret) {
                VCSAG_LOG_MSG("E", "Role-swap of the device group $self->{'name'} failed; error code is $ret", 53, "$self->{'name'}", "$ret");
                $self->_exit(0);
            }

            $count = $self->get_count();
            VCSAG_LOG_MSG("N", "issuing establish/resume of device group $self->{'name'}", 47, "$self->{'name'}");
            $ret = $self->run_cmd_chk_err("$symrdf $self->{'type'} $self->{'name'} $self->{'rdfgopt'} -noprompt -i 5 -c $count establish", $func);
            if($ret) {
                VCSAG_LOG_MSG("E", "Establish=link for the device group $self->{'name'} failed; error code is $ret", 54, "$self->{'name'}", "$ret");
                $self->_exit(0);
            }

            $count = $self->get_count();
            if ($disabled_consistency) {
                VCSAG_LOG_MSG("N", "re-enable consistency for device group $self->{'name'}", 48, "$self->{'name'}");
                $ret = run_cmd("$symrdf $self->{'type'} $self->{'name'} $self->{'rdfgopt'} -noprompt -i 5 -c $count enable >$DEVNULL", 0, 1, $func);
                if ($ret) {
                    VCSAG_LOG_MSG("N", "re-enabling consistency on device group $self->{'name'} failed. Administrative intervention required.", 49, "$self->{'name'}");
                    $self->_exit(0);
                }
            }
            sleep(3);
            $self->get_state();
        }
    }
    if ($self->{'state'} eq "RW") {
        VCSAG_LOG_MSG("N", "failover of devices in group $self->{'name'} succeeded", 23, "$self->{'name'}");
        $self->vxvm_scan();
        $self->create_lockfile();
        if ($self->{'is_vmware'}) {
            $self->vmware_online();
        } else { 
            $self->cvm_import();
        }
        $self->reload_devicemap();
        $self->_exit(0);
    } else {
        VCSAG_LOG_MSG("E", "failover of group $self->{'name'} failed, devices are not read-write enabled", 80, "$self->{'name'}");
        $self->_exit(0);
    }
}

sub create_device_files() {
    my($self, $fileName, $LsymmID, $RsymmID, $RDFG) = @_;
    my ($i, $j, $k, @localDevices, @remoteDevices, $retVal) ;
    my ($keyLocal, $keyRemote);

    $retVal = 1;

    $keyLocal = "Local" . $RDFG . $LsymmID;
    $keyRemote = "Remote" . $RDFG . $RsymmID;

    foreach $k (keys (%{$self->{ARRAYDEVS}})) {
        foreach $j (keys (%{$self->{ARRAYDEVS}{$k}})) {
            if ($k eq $keyLocal) {
                push(@localDevices, $self->{ARRAYDEVS}{$k}{$j});
            } elsif ($k eq $keyRemote) {
                push(@remoteDevices, $self->{ARRAYDEVS}{$k}{$j});
            }
        }
    }
    
    if ($^O =~ /MSWin32/i) {
	    $$fileName = "$vcs_log" . "\\lock\\.deviceFile" . "_" . $self->{'name'} . "_" . $LsymmID . "_" . $RDFG . "_" . "$self->{'resname'}";
    } else {
	    $$fileName = "$vcs_log" . "/lock/.deviceFile" . "_" . $self->{'name'} . "_" . $LsymmID . "_" . $RDFG . "_" . "$self->{'resname'}";
    }
    open DEVICEFILE, ">$$fileName" || die ("Cannot open $$fileName file!\n");
    for ($i=0; $i<$#localDevices + 1; $i++) {
        print DEVICEFILE "$localDevices[$i] $remoteDevices[$i]\n";
    }
    close DEVICEFILE;
    $retVal = 0;

    return $retVal;
}

sub remove_device_files() {
    my($self, $fileName) = @_;
    my $retVal = 0;

    unlink $fileName;
    return $retVal;
}

sub vmware_online {
	# Make sure that /proc/vmware/config/LVM/EnableResignature and
	# /proc/vmware/config/LVM/DisallowSnapshotLun both are set to 0
	# rescanning will be done by parent VMFS resource.
    my $cmd = "esxcfg-advcfg -s 0 /LVM/EnableResignature";
    system ($cmd);
    $cmd = "esxcfg-advcfg -s 0 /LVM/DisallowSnapshotLun";
    system ($cmd);

    return 0;
}

sub get_lockfilename {
    my ($self) = shift;
    my $typ;
    my $lock_file;
    my $lock_dir = ""; 

    if (!defined($self->{'resource_type'}) || 
            (length($self->{'resource_type'}) == 0)) {

        $typ = `$ag_i18n_inc::HARES -value $self->{'resname'} Type`;
        chomp $typ;
    } else {
        $typ = $self->{'resource_type'};
    }

    if ($^O =~ /MSWin32/i) {
        $lock_dir = $ag_i18n_inc::VCSLOG . "\\lock\\" . $typ . "\\";
    } else {
        $lock_dir = $ag_i18n_inc::VCSLOG . "/lock/" . $typ . "/";
    }

    $lock_file = $lock_dir . ".VCS_" . $typ . "_" . $self->{'resname'} . 
        "_Online_Lockfile.DO-NOT-REMOVE";

    return $lock_file;
}


sub create_lockfile {
    my ($self) = shift;
    my ($lock_file) = $self->get_lockfilename();

    my (undef, $lock_dir, undef) = fileparse($lock_file);

    if (! -d $lock_dir) {
        mkdir  $lock_dir or 
            VCSAG_LOG_MSG("E", "Could not create lock directory $lock_dir",
                    1001, "$lock_dir");
    }

    open(LFH, ">$lock_file") or 
        VCSAG_LOG_MSG("E", "Could not create file: $lock_file", 1002,
                "$lock_file");

    close(LFH);
}


sub remove_lockfile {
    my($self) = shift;
    my ($lock_file) = $self->get_lockfilename();

    if (-f "$lock_file") {
        unlink("$lock_file") or 
            VCSAG_LOG_MSG("E", "Could not delete file: $lock_file", 1003,
                    "$lock_file");
    }
}


sub exists_lockfile {
    my ($self) = shift;
    my $retval = 0;
    my ($lock_file) = $self->get_lockfilename();

    if (-f "$lock_file") {
        $retval = 1;
    }
    return $retval;
}

#
# Return 1 if a parent resource of a parent service group is ONLINE
#

sub is_parent_online {
    my ($self) = shift;
    my ($p_res, $ret, $tmpstr, $sys);
    my @p_res_list = ();
    my $is_online = 0;

    $tmpstr = `$ag_i18n_inc::HASYS -nodeid`;
    chomp $tmpstr;
    $tmpstr =~ s/^.*Node name:[\s+](.*)[\s+]Node ID:.*$/$1/ ;
    $self->{'nodename'} = $sys = $tmpstr;

# Are there any resources that depend on this resource
    @p_res_list = $self->get_res_deps();

    if ($#p_res_list >= 0) { # Parent resources were found
        foreach $p_res (@p_res_list) {

            $ret = system("$ag_i18n_inc::HARES -wait $p_res Probed 1 "
                    ." -sys $sys -time 120 2>$DEVNULL");
            if ($ret == 0) {
                my ($st) = split('\n', `$ag_i18n_inc::HARES -state $p_res -sys $sys`);
                chomp $st;
                if ($st =~ /ONLINE/) {
                    $is_online = 1;
                    last;
                }
            }
        }
    } else { # No parent resources
        if ($self->is_parent_sg_res_online()) {
            $is_online = 1;
        }
    }

    return $is_online;
}

#
# Return 1 if at least one OnOff/OnOnly resource from a parent service
# group is online.
#

sub is_parent_sg_res_online {
    my ($self) = shift;
    my ($sg_name, $p_sg_name, $line, $ret);
    my @lines = ();
    my @fields = ();
    my @valid_dep_lines = ();
    my $is_online = 0;
    my $sys = $self->{'nodename'};

#
# Get parent service groups that have an online local/online global
# dependency to our service group
#

    $sg_name = `$ag_i18n_inc::HARES -value $self->{'resname'} Group`;
    chomp $sg_name;

    (@lines) = split('\n', `$ag_i18n_inc::HAGRP -dep $sg_name`);
    chomp @lines;

# VCS specific code --->
    (@valid_dep_lines) = grep (/online[\s+](local|global)/, @lines );

    foreach $line (@valid_dep_lines) {
        (@fields) = split('\s+', $line);

        if ($fields[1] eq $sg_name) {
            $p_sg_name = $fields[0];
        } else {
            next;
        }

		#
		# If at least one OnOff or OnOnly probed resource in the ParentSG
		# has gone online, then the SRDF resource can be online as well
		#

        $ret = system("$ag_i18n_inc::HAGRP -wait $p_sg_name Probed 1 -sys $sys"
                . " -time 120 2>$DEVNULL");

        if ($ret == 0) {
            my @p_on_res_list = $self->get_res_list($p_sg_name, "ONLINE");

			#
			# At least 1 online resource was found. Note: $#X == 0
			# implies @X has 1 element
			#
            if ($#p_on_res_list >= 0){

				#
				# We found at least one ONLINE resource. If it is of type 
				# OnOff or OnOnly, we can continue to remain ONLINE
				#   
                foreach my $p_res (@p_on_res_list) {
                    my $p_type = `$ag_i18n_inc::HARES -value $p_res Type`;
                    chomp $p_type;
                    my $p_oprn_type = `$ag_i18n_inc::HATYPE -value $p_type Operations`;
                    chomp $p_oprn_type;
                    if ($p_oprn_type =~ /None/i ) {
                        next;
                    }

                    $is_online = 1;
                    last;
                }
            }
        }
        last if ($is_online == 1);
    }

    return $is_online;
}

#
# Returns the names of resources that belong to a given group and
# are in particular state
#
sub get_res_list {
    my ($self) = shift;
    my ($sg_name, $res_state) = @_;
    my @retarr = ();
    my @tmparr = ();
    my ($ret, $nodename, $resname);
    my $func = "get_res_list";

    $ret = run_cmd("$ag_i18n_inc::HARES -list Group=$sg_name State=$res_state", 
            \@tmparr, 1, $func);

    if (!$ret) {
        foreach (@tmparr) {
            /^#/ && next;
            ($resname, $nodename)  = split;

            if ((!grep(/^${resname}$/, @retarr)) &&
                    ($nodename =~ /$self->{'nodename'}/)) {

                push (@retarr, $resname);
            }
        }
    }
    return @retarr;
}


#
# Returns the names of parent resources
#
sub get_res_deps {
    my ($self) = shift;

    my ($sg_name, $res_state) = @_;
    my @retarr = ();
    my @tmparr = ();
    my ($ret, $grpname, $p_resname, $c_resname);
    my $func = "get_res_deps";

    $ret = run_cmd("$ag_i18n_inc::HARES -dep $self->{'resname'}", \@tmparr, 1, $func);

    if (!$ret) {
        foreach (@tmparr) {
            /^#/ && next;
            ($grpname, $p_resname, $c_resname)  = split;

			#
			# Note: with VCSOne, the 'resname' would be <grpname>.<resname>
			# whereas the $c_resname would only have the trailing <resname> 
			# part
			#

            if ((!grep(/^${p_resname}$/, @retarr)) && 
                    ( $c_resname =~ /$self->{'resname'}$/ )) {
                push (@retarr, $p_resname);
            }
        }
    }

    return @retarr;
}


sub run_cmd {
    my ($cmd, $output_ref, $ignore_failure, $func) = @_;
    my $retval = 0;

    VCSAG_LOGDBG_MSG(2, "$func: Issuing command: [$cmd]");
    my $typ = ref($output_ref);
    if (!$typ) { # output is not a ref. So caller is ot interested in output.
        system($cmd);
        return $?;
    }

    my $cmd_txt = $cmd;
    if (1 == $ignore_failure) {
        $cmd_txt .= " 2>$DEVNULL";
    }
    if ($typ eq "ARRAY") {
        @{$output_ref} = `$cmd_txt`;
        chomp @{$output_ref};
    } else {
        $$output_ref = `$cmd_txt`;
        chomp ${$output_ref};
    }

    $retval = $?;
    if ($retval != 0) {
        unless (1 == $ignore_failure ) {
            $cmd_txt =~ s/\"//g;
            VCSAG_LOG_MSG("W", "Command: \'$cmd_txt\' failed !", 1007, "$cmd_txt");
        }
    }
    return $retval;
}

sub __hack_se7_0_X_bug1 {
    my ($self, $recv_ret, @out) = @_;
    my $pattern = "The device is already in the requested RDF state";
    my $ret;

    VCSAG_LOGDBG_MSG(1, "Entering __hack_se7_0_X_bug1 ");

    if ($recv_ret != 1) {
        $ret = $recv_ret;
        VCSAG_LOGDBG_MSG(1, "__hack_se7_0_X_bug1 :: VALID RET");
    } else {
        chomp @out;
        $ret = 18 if (!grep(/$pattern/, @out)) ;
        VCSAG_LOGDBG_MSG(1, "__hack_se7_0_X_bug1 :: 7.0.X BUG Found");
    }
    return $ret;
}


# returns '-cg' if group is a composite group, '-g' otherwise
sub get_cg_options {
    my ($self) = shift;
    return ($self->{'iscompositegroup'}) ? "-cg" : "-g";
}

sub get_symapi_servers
{
    my ($resname, $symhome) = @_;
    return if ($OS eq "Windows NT");
    my $func = "get_symapi_servers";

	#
	# Get the list of SYMAPI_SERVERs if populated.
	#
    VCSAG_LOGDBG_MSG(1, "get_symapi_servers(): reading value of SymapiServers");
    my $cmd_op;
    run_cmd("$ag_i18n_inc::HARES -value $resname SymapiServers", \$cmd_op, 1, $func);
    chomp $cmd_op;
    my @symapi_servers = split(/\s+/, $cmd_op);
    if (@symapi_servers) {
        my $symapi_server = "";
        foreach my $sym_serv (@symapi_servers) {
            VCSAG_LOGDBG_MSG(1, "Checking symapi server: [$sym_serv]");
            $ENV{SYMCLI_CONNECT} = $sym_serv;
            run_cmd("$symhome/bin/symcfg list 2>/dev/null", \$cmd_op, 1, $func);
            VCSAG_LOGDBG_MSG(2, "cmd returned: [$cmd_op]");
            my @local_arrays = grep(/Local/, split(/\n/, $cmd_op));
            if (!@local_arrays) {
                VCSAG_LOGDBG_MSG(1, "Invalid entry.");
                delete $ENV{SYMCLI_CONNECT};
                next;
            }
            VCSAG_LOGDBG_MSG(1, "Agent will temporarily use [$sym_serv] as the remote symapi server for all operations in this entry point.");
            $symapi_server = $sym_serv;
            last;
        }
        if (!$symapi_server) {
            VCSAG_LOG_MSG("E", "None of the Symapi Servers specified for this resource are currently available. Please check whether the correct Symapi Servers are configured in the the symapi configuration options file. The resource will not go online until then.", 1009);
            exit(0);
        }
    }
    else {
        VCSAG_LOGDBG_MSG(1, "No symapi server configured. Using local mode.");
    }
}

sub get_rdf_groups
{
    my ($self, $lines_ref) = @_;
    my @lines = @$lines_ref;

    VCSAG_LOGDBG_MSG(1, "get_rdf_groups: Entered.");
    my $rdfs = ();
    my $rdfinfo = {};
    my $old_rdfdata = 0;
    my $r2_rdfg = "";
    my $r2_rdfg_count = 0;
    my $preferred_site_match = 0;
    foreach my $line (@lines) {
        chomp $line;
        next if ($line =~ /^$/);
        if ($line =~ /Device Group RDF Information/) {
            if ($old_rdfdata) {
                $rdfs->{$rdfinfo->{'symid'}} = $rdfinfo;
                if ($rdfinfo->{'type'} eq "R2") { # Assuming that the local device will be R2 in only one rdfg.
                    $self->{'preferredsite'} = $rdfinfo->{'symid'};
                }
                $rdfinfo = {};
            }
            $old_rdfdata = 1;
            next;
        }
        if ($line =~ /RDF Type\s+:\s+(\S+)/) {
            $rdfinfo->{'type'} = $1;
            next;
        }
        if ($line =~ /RDF \(RA\) Group Number\s+:\s+(\S+)/) {
            $rdfinfo->{'rdfg'} = $1;
            next;
        }
        if ($line =~ /Remote Symmetrix ID\s+:\s+(\S*)/) {
            $rdfinfo->{'symid'} = $1;
            next;
        }
        if ($line =~ /RDF Mode\s+:\s+(\S*)/) {
            $rdfinfo->{'mode'} = $1;
            next;
        }
        if ($line =~ /RDF Pair Configuration\s+:\s+(\S*)/) {
            $rdfinfo->{'pairconfig'} = $1;
            next;
        }
        if ($line =~ /RDF STAR Mode\s+:\s+(\S*)/) {
            $rdfinfo->{'starmode'} = $1;
            next;
        }
        if ($line =~ /^\s+Device RDF State\s+: (.*)/) {
            $rdfinfo->{'rdfstate'} = $1;
            $rdfinfo->{'rdfstate'} =~ s/\s+/ /g;
            next;
        }
        if ($line =~ /Remote Device RDF State\s+: (.*)/) {
            $rdfinfo->{'remotestate'} = $1;
            $rdfinfo->{'remotestate'} =~ s/\s+/ /g;
            next;
        }
        if ($line =~ /RDF Pair State .*: (.*)/) {
            $rdfinfo->{'pairstate'} = $1;
            $rdfinfo->{'pairstate'} =~ s/\s+/ /g;
            next;
        }
    }
    if ($old_rdfdata) {
        $rdfs->{$rdfinfo->{'symid'}} = $rdfinfo;
        if ($rdfinfo->{'type'} eq "R2") { # Assuming that the local device will be R2 in only one rdfg.
            $self->{'preferredsite'} = $rdfinfo->{'symid'};
        }
        $rdfinfo = {};
    }
    if ($self->{'preferredsite'}) {
        VCSAG_LOGDBG_MSG(1, "get_rdf_groups: Found R2 RDFG.");
    } else {
        VCSAG_LOGDBG_MSG(1, "get_rdf_groups: No R2 RDFG.");
    }
    $self->{'rdfg'} = $rdfs;
    $self->{'preferredrdfg'} = $rdfs->{$self->{'preferredsite'}}->{'rdfg'};
    $self->{'rdfgopt'} = "-rdfg $self->{'preferredrdfg'}" if ($self->{'preferredrdfg'});
}

sub get_symms {

    my ($self, $lsymms_ref, $rsymms_ref) = @_;
    my @tmparr = ();
    my @lsid_lines;
    my @rsid_lines;
    my @lines;
    my $func = "get_symms";
    my $mode;
    $self->{'concurrent'} = 0;

    SRDFAgent::get_symapi_servers($self->{'resname'}, $self->{'home'});

    @{$lsymms_ref} = ();
    @{$rsymms_ref} = ();

    my $showcmd = ($self->{'iscompositegroup'} != 0) ? 
        "$symcg show $self->{'name'}" : 
        "$symdg show $self->{'name'}";
    if (SRDFAgent::run_cmd("$showcmd", \@tmparr, 1, $func)) {
        $self->_exit(0);
    }

    @lines = grep (/Symmetrix ID/, @tmparr);
    if ($OS ne "Windows NT") {
        VCSAG_LOGDBG_MSG(1, "get_symms Checking for concurrent...");
        my @rdf_groups = grep(/Device Group RDF Information/, @tmparr);
        my $num_rdfg = @rdf_groups;
        if ($num_rdfg > 1) {
            $self->{'concurrent'} = 1;
            VCSAG_LOGDBG_MSG(1, "Concurrent SRDF setup detected.");
            $self->get_rdf_groups(\@tmparr); # Create a map of Remote SymIds to RA Group No. Also sets the preferred site rdfg no.
        } else {
            $self->{'concurrent'} = 0;
        }
    }
# Find the Group Type for faster decision making.
    VCSAG_LOGDBG_MSG(1, "get_symms Checking for fast failover...");
    my $std_context = 0;
    $self->{'dev_state'} = "";
    foreach my $line (@tmparr) {
        $line =~ s/^\s+//;
        $line =~ s/\s+$//;
        if ($line =~ /Group Type\s+:\s+(\S+)/) {
            $self->{'grp_type'} = $1;
            next;
        }
        if ($line =~ /Standard \(STD\) Devices/) {
            $std_context = 1;
            next;
        }
        if ($std_context) {
            if ($line =~ /\}/) {
                $std_context = 0;
                next;
            }
            my @tokens = split(/\s+/, $line);
            if ($#tokens == 5 && $tokens[$#tokens] =~ /^\d+$/) {
                if ($self->{'dev_state'} eq "") {
                    $self->{'dev_state'} = $tokens[4];
                } else {
                    if ($self->{'dev_state'} ne $tokens[4]) {
                        $self->{'dev_state'} = "mix";
                    }
                }
            }
        }
		if ($line =~ /RDF Mode\s+:\s+(\S+)/) {
			$mode = $1;
			next
		}
		# In case of concurrent setup SRDF agent can come online only on the Synchronous link
		if ($line =~ /RDF Pair State .*:\s+(.*)/) {
			if ( $self->{'concurrent'} == 1 && $mode eq "Synchronous") {
				$self->{'pairstate'} = $1;
				$self->{'pairstate'} =~ s/\s+/ /g;
			} elsif ( $self->{'concurrent'} == 0) {
				$self->{'pairstate'} = $1;
				$self->{'pairstate'} =~ s/\s+/ /g;
			}
		}
    }
#
# for composite groups, @lines contains lines of the form:
#   1) Symmetrix ID                                   : 000187400527
#      Remote Symmetrix ID    : 000187900584
# for rdf device groups, the @lines is of the form:
#   Symmetrix ID                                   : 000187400527
#      Remote Symmetrix ID    : 000187900584
#

    @lsid_lines =  grep(!/Remote Symmetrix ID/, @lines);
    @rsid_lines =  grep(/Remote Symmetrix ID/, @lines);

    if (scalar(@lsid_lines) <= 0 or scalar(@rsid_lines) <= 0) {
        $ENV{"LANG"}="$USERLANG";
        VCSAG_LOG_MSG("E", "Could not find any Symmetrix ID associated with group $self->{'name'}\n", 1008, "$self->{'name'}");
        $self->_exit(0);
    }

    foreach my $line (@lsid_lines) {
        my (undef, $sid) = split (/:/, $line);
# get rid of all whitespaces
        $sid =~ s/\s+//g;
        push (@{$lsymms_ref}, $sid) if (!grep(/$sid/, @{$lsymms_ref}));
    }

    foreach my $line (@rsid_lines) {
        my (undef, $sid) = split (/:/, $line);
# get rid of all whitespaces
        $sid =~ s/\s+//g;
        push (@{$rsymms_ref}, $sid) if (!grep(/$sid/, @{$rsymms_ref}));
    }

    return;
}

#
# Returns 0 if any command failed
# Returns 1 otherwise
#
sub _get_dynamic_devs {
    my ($self, $dyndevs_href, @symms) = @_;
    %{$dyndevs_href} = ();
    my $func = "_get_dynamic_devs";

    my ($ret);
    my @tmparr = ();

    foreach my $localsymm (@symms) {
        my @dyndevs = ();
        if (SRDFAgent::run_cmd("$symdev -dynamic list -sid $localsymm", \@tmparr, 1, $func)) {
            return 0 ;
        }
        foreach my $line (@tmparr) {
            if ($line =~ /RDF/) {
                my (@fields) = split('\s+', $line);
                push(@dyndevs, $fields[0]);
            }
        }
        $dyndevs_href->{$localsymm} = [ @dyndevs ];
        if ($#dyndevs == -1) {
            return 0;
        }

    }
    return 1;
}

#
# Returns 0 if any command failed
# Returns 1 otherwise
#
sub _get_grp_devs {
    my ($self,$type, $grpdevs_href, @local_symms) = @_;
    %{$grpdevs_href} = ();
    my $func = "_get_grp_devs";

    my @fields = ();
    my @tmparr = ();
    my $cmd;
	my $modepat= qr/[SAD][X\.]/;
    if($self->{'symcli_ver'}<9.0) {
              $modepat= qr/[SAC]\./;
	}
    foreach my $localsymm (@local_symms) {
        my @grpdevs = ();

        $cmd = "$symrdf $type $self->{'name'} query $self->{'rdfgopt'}";
#
# Composite groups can span multiple symmetrix arrays. We need to find
# our devices on each array
#
        if ($self->{'iscompositegroup'} != 0) {
            $cmd .= " -sid $localsymm";
        }

        if (SRDFAgent::run_cmd("$cmd", \@tmparr, 1, $func)) {
            return 0 ;
        }

# Consider only lines that contain  /[SAC]\./ or /[SAD][X\.]/  eg: S. or A. etc
        foreach my $line (@tmparr) {
            if ($line =~ /Group Number/) {
                next;
            }
            if ($line =~ /^Remote Symmetrix ID/) {
                next;
            }
            if ($line =~ /^Symmetrix ID/ or $line =~ /^DG's Symmetrix ID/) {
                next;
            }
            if ($line =~ /Source/) {
                next;
            }
            if (!($line =~/$modepat/)) {
		next;
            }
            if ($line =~ /Device Group/) {
                next;
            }
            (@fields) = split('\s+', $line);
            push(@grpdevs, $fields[1]);
        }
        $grpdevs_href->{$localsymm} = [ @grpdevs ];

        if ($#grpdevs == -1) {
            return 0;
        }

    }

    return 1;
}
# Check if all our devices are dynamic devices
# Returns undef if any command failed, 
# 1 if true, 0 on false
#
sub is_dynamic_grp_devs {
    my ($self, $type, @local_symms) = @_;

    my $retval = 0;
    my $found = 0;
    my %dyndevs_hash = (); 
    my %grpdevs_hash = ();

    if(!($self->_get_grp_devs($type, \%grpdevs_hash, @local_symms)) ||
            !($self->_get_dynamic_devs(\%dyndevs_hash, @local_symms))) {
		VCSAG_LOGDBG_MSG(1, "Devices in the group are not dynamic");
        	return 0;
    }

    foreach my $localsymm (@local_symms) {
        my @grpdevs = @{ $grpdevs_hash{$localsymm} };
        my @dyndevs = @{ $dyndevs_hash{$localsymm} };

# Trivial case wherein either list is empty
        if (($#grpdevs == -1) || ($#dyndevs == -1)) {
           	VCSAG_LOGDBG_MSG(1, "This is a trivial case wherein either list of group or dynamic devices is empty");
		return 0;
        }

		#
		# If any device from the group does not exist in the 
		# dynamic devices list, return 0;
		#

        foreach my $grpdev (@grpdevs) {
            if (!grep(/^$grpdev$/, @dyndevs)) {
  		     VCSAG_LOGDBG_MSG(1, "No device from the group is present in the dynamic devices list"); 
	             return 0;
            }
        }
    }
    VCSAG_LOGDBG_MSG(3, "Devices in the group are dynamic type");
    return 1;
}

#
# check that all connected arrays are at or above the supported microcode level
# returns 1 if all arrays are at/above supported microcode level, 0 otherwise 
#
sub is_mcode_version_OK {
    my ($self, $lsymms_ref) = @_;
    my $func = "is_mcode_version_OK";

    my @fields = ();
    my @tmparr = ();

    if (SRDFAgent::run_cmd("$symcfg list", \@tmparr, 1, $func)) {
        	return undef;
    }

	#
	# Create a string of the form: 
	#   "<local1 sid>|<local2 sid>|<remote1 sid>|<remote2 sid>"
	#

    my $srchpattern = @{$lsymms_ref};
    my @lines = grep (/$srchpattern/, @tmparr);

	#
	# Each line is of the form:
	#  '   000187400527 Local   DMX1000S  5671   16384  37 3577'
    foreach my $l (@lines) {
        (@fields) = split('\s+', $l);

        if ($fields[4] < SUPPORTED_MCODE_VERSION) {
       		VCSAG_LOGDBG_MSG(1,"mcode version is lesser than expected");
        	return 0; 
        }
    }
    VCSAG_LOGDBG_MSG(1,"mcode version is ok");
    return 1;
}

#
# Return 1 if Dynamic RDF config is enabled for the arrays; 0 otherwise
#
sub is_symmstate_dynamic {
    my ($self,@local_symms) = @_;

    my $retval = 0;
    my @tmparr = ();
    my $drc_state_cnt = 0;
    my $func = "is_symmstate_dynamic";

    foreach my $localsymm (@local_symms) {
        if (SRDFAgent::run_cmd("$symcfg -sid $localsymm list -v", \@tmparr, 1, $func)) {
	    VCSAG_LOGDBG_MSG(1,"Dynamic RDF config is not enabled");
            return 0;
        }
        if (grep (/\s+Dynamic RDF Configuration State\s+:\s+Enabled/, @tmparr)){
            $drc_state_cnt++ ;
        }
    }

    if ( $drc_state_cnt == scalar(@local_symms)){
        $retval = 1;
	VCSAG_LOGDBG_MSG(1,"Dynamic RDF config is enabled");
    }

    return $retval;
}

#
# CLI versions supported are 6.3 and above
# Returns 1 on success, 0 on error
#
sub is_CLI_version_OK {
	my ($self) = shift;
	my @tmparr = ();
	my $func = "is_CLI_version_OK";


# To workaround the issue in perl on Windows, run the command with some
# argument. backticks fail when the ShortPath contains special characters. 
# But strangely succeeds if we provide an argument.
	if (SRDFAgent::run_cmd("$symcli -v", \@tmparr, 1, $func)) {
		return 0;
	}

	$self->extract_CLI_version(\@tmparr);

	if(($self->{'symcli_ver'} =~ /6\.[3-9]/) || ($self->{'symcli_ver'} =~ /[7-9]\.[0-9]/)) {
		VCSAG_LOGDBG_MSG(1,"symcli version is within supported range");
		return 1;
	}
	else {
		$ENV{"LANG"}="$USERLANG";
		my $name = $self->{'name'};
		VCSAG_LOG_MSG("N", "Group $name is enabled for role swap on failover, but SYMCLI 6.3 or above is not installed; devices will be treated as normal devices", 60017 ,"$name");
		$ENV{"LANG"}="C";
		return 0;
	}
}

sub lock_res {
    my($self) = shift;
    my ($resname) = $self->{'resname'};

    my ($grp) = split('\n', `$ag_i18n_inc::HARES -value $self->{'resname'} Group`);
    my ($para) = split('\n', `$ag_i18n_inc::HAGRP -value $grp Parallel`);
    return 0 if ($para != 1);

    my ($lockholder) = split('\n', `$ag_i18n_inc::HARES -value $resname VCSResLock`);
    my (@locsysarr) = split('\s+', `LC_ALL=C; export LC_ALL; $ag_i18n_inc::HASYS -nodeid`);
    my ($thisnode) = $locsysarr[5];
    if ($lockholder eq "") {
        `$ag_i18n_inc::HARES -lock $resname > /dev/null 2>&1`;
        my $retval = $?;
        return 0 if ($retval == 0);
        ($lockholder) = split('\n', `$vcs_home/bin/hares -value $resname VCSResLock`);
    }
    return $lockholder if ($lockholder ne $thisnode);
    return 0;
}

sub unlock_res {
    my($self) = shift;

    my ($grp) = split('\n', `$ag_i18n_inc::HARES -value $self->{'resname'} Group`);
    my ($para) = split('\n', `$ag_i18n_inc::HAGRP -value $grp Parallel`);
    return 0 if ($para != 1);

    `$ag_i18n_inc::HARES -unlock $self->{'resname'} > /dev/null 2>&1`;
}

#Get disk list and call vxdisk scandisks device=<device list>
sub get_devices {
	my ($self) = shift;
	my $selctivescan = shift;
	my @disks;
        my $ret = 0;
	my $func = "get_devices";
	my (@diskdetarr);
	my $searchdev = "DEV\\d";
	my $grepstring = "'";
	$ret = $self->get_disks(\@disks, $searchdev);
        if ($ret){
		VCSAG_LOG_MSG("W", "Could not get devices from DG: $self->{'name'}", 1114, "$self->{'name'}");
                return 1;
        }
	my $flag = 1;
	for my $disk (@disks) {

		if($flag){
			$grepstring .= $disk;
			$flag = 0;
		}
		else {
			$grepstring .= "|";
			$grepstring .= $disk;
		}
	}
	if($flag){
		VCSAG_LOG_MSG("W", "Could not get single device id from DG: $self->{'name'}", 1113, "$self->{'name'}");
		return 1;
	}
	$grepstring .= "'";
        VCSAG_LOG_MSG("I", "CLI:$VXDISK path | egrep -i $grepstring", 1118, "$VXDISK", "$grepstring");
	$ret = run_cmd("$VXDISK path | egrep -i $grepstring", \@diskdetarr, 0, $func);
	if($ret != 0){
		VCSAG_LOG_MSG("W", "Failed to list STD devices in DG: $self->{'name'}", 1115, "$self->{'name'}");
		return 1;
	}

	my $scancmd = "$VXDISK scandisks device=";
	my $diskname;
	$flag = 1;
	for my $lines (@diskdetarr){
		my (@fields) = split('\s+', $lines);
		$diskname = $fields[0];
		VCSAG_LOGDBG_MSG(3,"Extracted disk = $diskname");
		if($flag eq 1) {
			$scancmd .= $diskname ;
			$flag = 0;
		}
		else {
			$scancmd .= ",";
			$scancmd .= $diskname ;
		}
	}
	$$selctivescan = $scancmd;
	return 0;
}


sub get_disks {
        my ($self) = shift;
        my $disk_ref = shift;
        my $searchdev = shift;
        my $ret = 0;
        my $ArraySno= " ";
        my (@disks, @lines);
	my $osType = $^O;
    my (@syminq_output, @vxdiskpath_op);

    #Store syminq command output
    if ($osType !~ /^MSWin/i) {
        (@syminq_output) = split('\n', `$syminq -pdevfile`);
        (@vxdiskpath_op) = split('\n', `$VXDISK path`);
    }
        #
        # First get the R2 device names from sym
        #
        if($self->{'iscompositegroup'} == 0) {
		VCSAG_LOG_MSG("I", "calling $symdg show $self->{'name'}", 1117, "$symdg", "$self->{'name'}");
                (@lines) = split('\n', `$symdg show $self->{'name'}`);
        } else {
		VCSAG_LOG_MSG("I", "calling $symcg show $self->{'name'}", 1119,"$symcg","$self->{'name'}");
                (@lines) = split('\n', `$symcg show $self->{'name'}`);
        }
        foreach my $l (@lines) {
                if ($l =~ /Symmetrix\sID/) {
                    if (!($l =~ /Remote\sSymmetrix\sID/)) {
                                my (@fields2) = split (':',$l);
                                $ArraySno = $fields2[1];
                 #Trim Array Serial Number
                                $ArraySno =~ s/^\s+//;
                                $ArraySno =~ s/\s+$//;
                    }
                }
                elsif ($l !~ /$searchdev/) { next; }
                my (@fields) = split('\s+', $l);
                if (!($fields[1] =~ /^$searchdev/)) { next;}
        my ($vxdiskname, $symdevid);
        if ($osType =~ /^MSWin/i) {
            $vxdiskname = $fields[2];
            $vxdiskname =~ s/.*\///;
            $vxdiskname =~ s/\\\\.\\PHYSICALDRIVE/Harddisk/;
        } else {
		$symdevid = $fields[3];
		VCSAG_LOGDBG_MSG(2, " symdevid = $symdevid");
		$vxdiskname = findvxdiskname($ArraySno,$symdevid,\@syminq_output,\@vxdiskpath_op);
                VCSAG_LOGDBG_MSG(1, "vxdiskname: $vxdiskname");
        }
                if ($vxdiskname ne 'none found') {
            push(@{$disk_ref}, $vxdiskname);
        } else {
		VCSAG_LOG_MSG("W", "Failed to get the node name for symdev:  $symdevid ", 1120, "$symdevid");
		return 1;
                }
        }
        return 0;
}


sub findvxdiskname
{
        my ($ArraySno) = shift;
        my ($devid) = shift;
        my $cmdoutput_ref = shift;
        my $vxdiskpath_op_ref = shift;
        my @cmdoutput = @$cmdoutput_ref;
        my @vxdiskpath_op = @$vxdiskpath_op_ref;

        foreach my $l (@cmdoutput) {
                if ($l =~ /$devid/ ){
                        my (@fields) = split('\s+', $l);
                        my $ArraySno2 = $fields[0];
                        my $devpath = $fields[1];
                        my $devid1 = $fields[2];
                        if ($ArraySno2  eq $ArraySno)
                        {
                                my ($OSname)  = (POSIX::uname())[0];
                                if ($devpath =~ /^\Q$VXPATH\E/)
                                {
                                        my (@pathfields) = split('/',$devpath);
                                        if ( $OSname eq "SunOS" )   # Snip trailing "s2"  for Solaris
                                        {
                                                $pathfields[4] =~ s/s2$//;
                                        }
                                        VCSAG_LOGDBG_MSG(1, "findvxdiskname pathfields[4] = $pathfields[4]");
                                        return $pathfields[4];
                                } else {
                                        VCSAG_LOGDBG_MSG(1, "findvxdiskname: [$devpath]. Trying to find vxdisk name.");
                                        if ($devpath =~ /(\w+)$/) {
                                                my $osh = $1;
						VCSAG_LOGDBG_MSG(3, "findvxdiskname diskname = $osh");
                                                 if ( $OSname eq "SunOS" )   # Snip trailing "s2"  for Solaris
                                                {
                                                        $osh =~ s/s2$//;
                                                } elsif ($OSname eq "AIX") {
                                                        $osh =~ s/^r//;
                                                }
						VCSAG_LOGDBG_MSG(3, "findvxdiskname after substitution diskname = $osh");
                                                my @lines = grep(/^$osh\W+/, @vxdiskpath_op);
                                                my $num = @lines;
						VCSAG_LOGDBG_MSG(1, "findvxdiskname number of line return = $num");
                                                if ($num == 0) {
                                                # Additional grep required because on solaris 10 name of the device has s2 as a suffixx
                                                        @lines = grep(/^$osh/, @vxdiskpath_op);
                                                        $num = @lines;
                                                        if ($num == 1) {
                                                                $lines[0] =~ /^$osh\w+\s+(\w+)/;
                                                                my $vxdisk = $1;
                                                                return $1;
                                                        }
                                                }
                                                if ($num != 1) {
                                                        VCSAG_LOGDBG_MSG(1, "findvxdiskname: Multiple matches!");
                                                        next;
                                                }
                                                $lines[0] =~ /^$osh\s+(\w+)/;
                                                my $vxdisk = $1;
                                                return $vxdisk;
                                        }
                                }
            }
        }
    }
    return "none found";
}

#################################################################
#sub will check DMP pkg level and prevent execution
#of vxdisk rm command having following pkg level.
#1)     6.2+
#2)     6.0.5.*
#3)     6.1.1.*
#Return value:0 if expected pkg level, else 1.
#Sample o/p line from lslpp
#VRTSvxvm    6.1.0.0    C     F    Veritas Volume Manager by
###################################################################

sub check_dmp_pkg_level {

	my $func = "check_dmp_pkg_level";
	my ($mjor, $minor, $third) = (0,0,0);
	my @lines;
	my @cmd_op = ();

	if(run_cmd("lslpp -L VRTSvxvm", \@cmd_op, 1, $func)) {
		VCSAG_LOG_MSG("N", "Failed to execute lslpp", 1120);
		return 1;
	}

	foreach my $line (@cmd_op) {
		if ($line =~ /VRTSvxvm.*?(\d+).(\d+).(\d+)/) {
			$mjor=$1;
			$minor=$2;
			$third=$3;
			VCSAG_LOG_MSG("I", "Vxvm pkg level Major:$mjor, Minor:$minor", 1120,"$mjor","$minor");
			VCSAG_LOGDBG_MSG(1, "check_dmp_pkg_level:build=$third");
			last;
		}
	}

	if($mjor > 6) {

		VCSAG_LOGDBG_MSG(1, "check_dmp_pkg_level:Expected vxvm verision found.");
		return 0;
	}

	if($mjor == 6) {
		if($minor >= 2) {
			VCSAG_LOGDBG_MSG(1, "check_dmp_pkg_level:Expected vxvm verision found.");
			return 0;
		}
		if(($minor == 0) && ($third =~ /^5/)) {  
			VCSAG_LOGDBG_MSG(1, "check_dmp_pkg_level:Expected vxvm verision found.");
			return 0;
		}

		if(($minor == 1) && ($third =~ /^1/)) {
			VCSAG_LOGDBG_MSG(1, "check_dmp_pkg_level:Expected vxvm verision found.");
			return 0;
		}

	}
	VCSAG_LOGDBG_MSG(1, "check_dmp_pkg_level:Expected vxvm verision not found.");
	return 1;
}


sub check_device_conf {
	my $self = shift;
	my $ret = 0;
	my $searchdev = "DEV\\d";

	my $number_devices = $self->get_device_count($searchdev);
	VCSAG_LOGDBG_MSG(3,"check_device_conf: number of devices detected[$number_devices] ");
        if($number_devices <= 0){
                VCSAG_LOG_MSG("W", "Invalid number of devices detected [$number_devices]", 60026,"$number_devices");
                return 1;
        }

	$ret = $self->get_remote_info($number_devices);
	if ($ret != 0) {
		VCSAG_LOG_MSG("E", "Failed to get RUNNING global Remote cluster node.", 60020);
		return 1;
	}

	return 0;
}

####################################################################
# Function :check_remote_site_conf 
# Desc     : Execute "validate_config" action on remote site
# Returns  : 1 Failed, 0 success.
#######################################################################

sub check_remote_site_conf {

	my $self = shift;
	my $number_devices = shift;
	my $rem_node = shift;
	my $rem_clus = shift;
	my $res_name = $self->{'resname'};
	my $cmd;
	my $ret;

	VCSAG_LOGDBG_MSG(3, "check_remote_site_conf enter");
	VCSAG_LOG_MSG("I", "Executing validate_config entry point on Remote node[$rem_node] Remote Clus [$rem_clus]", 60021, "$rem_node", "$rem_clus");
	if($rem_clus eq "") {
		$cmd = "$HARES -action $res_name validate_config -actionargs $number_devices -sys $rem_node";
	} else {
		$cmd = "$HARES -action $res_name validate_config -actionargs $number_devices -sys $rem_node -clust $rem_clus";
	}

	$ret = run_cmd("$cmd");
	if($ret){
		VCSAG_LOG_MSG("W", "Failed to check configuration on remote site or invalid configuration", 60022);
		return 1;
	}
	return 0;
}

####################################################################
# Function : get_remote_info
# Desc     : Retrieve the name of the remote node and cluster
# Returns  : 1 Failed, 0 success.
#######################################################################

sub get_remote_info {
	my $self = shift;
	my $res_id = $self->{'resname'};
	my $number_devices = shift;
	my $rem_node;
	my $rem_clus = "";
	my $ret = 0;
	my $ret_code = 1;
	open(RDSP, "$ag_i18n_inc::HARES -display $res_id |");
	if ($self->is_global_group($res_id)) {
		while (<RDSP>) {
			next unless /^$res_id\s+State\s/;
			next if /\slocalclus/;
			my @line = split();
			($rem_clus, $rem_node) = split(':', $line[2]);
			$ret = $self->is_node_running($rem_node);
			if ($ret == 1) {
				VCSAG_LOG_MSG("I", "Remote node $rem_node is not RUNNING. Looping for next node.", 60023, "$rem_node");
				next;
			}
			VCSAG_LOGDBG_MSG(3, "Running node name [$rem_node], cluster name[$rem_clus] ");
			$ret_code = $self->check_remote_site_conf($number_devices, $rem_node, $rem_clus);	
			if($ret_code) {
			VCSAG_LOGDBG_MSG(3, "check_remote_site_conf return failure. ");
				last;
			}
		}
# Make sure the pipe is empty
		while (<RDSP>) {next;}
		close RDSP;
	} else {
#changes for failover SG name
	$ret_code = $self->local_service_gp($number_devices);
	}

	return $ret_code;
}

#################################################################
# Function : local_service_gp 
# Desc     : Retrive the node name for failover SG 
# Returns  : 1 Failed, 0 success.
#################################################################

sub local_service_gp {
	my $self = shift;
	my $number_devices = shift;
	my $cmd;
	my $ret_code = 1;
	my $func = "local_service_gp";
	VCSAG_LOGDBG_MSG(3, "$func enter");
	my $out_put;

#get local system name
	$cmd = "$ag_i18n_inc::HASYS -nodeid";
	my $ret = run_cmd ($cmd, \$out_put, 1, $func);
	if($ret){
		VCSAG_LOG_MSG("W", "Failed to get local system name.", 60031);
		return 1;
	}

        my (@fields) = split ( '\s+' , $out_put);
        my $localsys = $fields[5];
	VCSAG_LOGDBG_MSG(1, "Local system name is : $localsys");
#get group name
	$cmd = "$ag_i18n_inc::HARES -value $self->{'resname'} Group";
	my $grp;
	$ret = run_cmd ($cmd, \$grp, 1, $func);
	if($ret){
		VCSAG_LOG_MSG("W", "Failed to get service group name.", 60032);
		return 1;
	}
#get group's system list
	$cmd = "$ag_i18n_inc::HAGRP -value $grp SystemList";
	my ($syslist) = split ('\n', `$cmd`);
	VCSAG_LOGDBG_MSG(1, "System list of group $grp : $syslist");

	my (@sysnames) = split('\s+', $syslist);

	foreach my $sysname (@sysnames) {
#exclude priority order
		if($sysname =~ /^[0-9]/) {
			next;
		}
#exclude local system
		if ($sysname =~ /$localsys/) {
			next;
		}
		VCSAG_LOGDBG_MSG(1, "System from group's system list :  $sysname");

		my $state;
		$cmd = "$ag_i18n_inc::HASYS -state $sysname";
		$ret = run_cmd ($cmd, \$state, 1, $func);
		if($ret || ($state ne "RUNNING") ){
			VCSAG_LOG_MSG("W", "Failed to get system state or node is not in running state.", 60033);
			return 1;
		}

		VCSAG_LOGDBG_MSG(3, "Running node name [$sysname]");
		$ret = $self->check_remote_site_conf($number_devices, $sysname);	
		if($ret) {
			VCSAG_LOGDBG_MSG(3, "check_remote_site_conf return failure. ");
			return 1;
		}

		$ret_code = 0;
	}

	return $ret_code;
} 

######################################################################
#Function : is_global_group
#check if configured service group(SG) is global group
# Returns  : 1 SG is global group, 0 SG is not global group.
#######################################################################

sub is_global_group {
	my $self = shift;
	my $res_name = $self->{'resname'};
	my $ret = 0;
	my $cmd = "$ag_i18n_inc::HARES -value $res_name Group";
	my $Grp = `$cmd`;
	chomp $Grp;
	$cmd = "$ag_i18n_inc::HAGRP -value $Grp  ClusterList";
	my $cluster_list = `$cmd`;
	chomp $cluster_list;

	if ( $cluster_list eq "" ) {
		VCSAG_LOG_MSG("I", "Service Group $Grp is not configured as global group.", 60024, "$Grp");
	} else {
		VCSAG_LOG_MSG("I", "Service Group $Grp is configured as global group.", 60025, "$Grp");
		$ret = 1;
	}
	return $ret;
}

#######################################################################
#Function :is_node_running
#check if node is running
# Returns  : 1 not running, 0 running.
#######################################################################

sub is_node_running {
	my $self = shift;
	my $rem_node = shift;
	my $ret = 1;
	open(RDSP1, "$ag_i18n_inc::HASYS -state |");
	while (<RDSP1>) {
		my @line = split();
		my ($Clus, $Node) = split(':', $line[0]);
		next if ($Node ne $rem_node);
		if ($line[2] eq "RUNNING") {
			$ret = 0;
		}
		last;
	}
#Make sure the pipe is empty
	while (<RDSP1>) {next;}
	close RDSP1;
	return $ret;
}

#######################################################################
#Function :compare_devices
#sub detect the number of search devices present in device
#group and compare it with number of devices received.
#return 0 if equal(success)
#return 1 if failure or devices number are different
#######################################################################

sub compare_devices {
	VCSAG_LOGDBG_MSG(3,"compare_devices entered");
	my ($self) = shift;
	my $searchdev = shift;
	my $total_lun = shift;
	my $ret = 1;
	if($total_lun <= 0){
		VCSAG_LOG_MSG("W", "Invalid number of devices received [$total_lun]", 60026,"$total_lun");
		return 1;
	}

	my $count = $self->get_device_count($searchdev);
	VCSAG_LOG_MSG("I", "Number of devices received [$total_lun], Number of devices detected[$count]", 60027,"$total_lun", [$count]);
	if($count == $total_lun) {
		$ret = 0;
	}
	return $ret;
}

####################################################################
# Function :get_device_count 
# Desc     : Execute symdg show / symcg show symcli and count the devices
# based on provided search string
#######################################################################

sub get_device_count {
	my $self = shift;
	my $searchdev = shift;
	my (@lines);
	my $count = 0;
	VCSAG_LOGDBG_MSG(3,"get_device_count entere");

	if($self->{'iscompositegroup'} == 0) {
		VCSAG_LOG_MSG("I", "calling $symdg show $self->{'name'}", 60028, "$symdg", "$self->{'name'}");
		(@lines) = split('\n', `$symdg show $self->{'name'}`);
	} else {
		VCSAG_LOG_MSG("I", "calling $symcg show $self->{'name'}", 60029,"$symcg","$self->{'name'}");
		(@lines) = split('\n', `$symcg show $self->{'name'}`);
	}

	foreach my $l (@lines) {
		if ($l !~ /$searchdev/) { next; }
		my (@fields) = split('\s+', $l);
		if (!($fields[1] =~ /^$searchdev/)) { next;}
		VCSAG_LOGDBG_MSG(3,"search device name = $fields[1]");
		$count++;
	}
	VCSAG_LOGDBG_MSG(3,"get_device_count return value[$count]");
	return $count;
}
sub extract_CLI_version{
	my ($self) = shift;
	my ($arrref) = @_;
	my @tmparr = @{$arrref};
	my $version = "";
	my $pattern= qr/(\d.\d.\d)/;
	foreach(@tmparr) {
		my $out1 = "";
		$out1 = $_;
		if(($out1 =~ /Symmetrix Command/) && ($out1 =~ /$pattern/)) {
			$self->{'symcli_ver'}=$1;	
			VCSAG_LOG_MSG("I", "Symcli version is: $self->{'symcli_ver'}", 60035, "$self->{'symcli_ver'}");
			last;
		}	
	} 
}
sub remove_space {
	my ($attribute) = @_;
	$attribute =~s/^\s+//;
	$attribute =~s/\s+$//;
	return $attribute;
}
sub set_db_file {
	my ($symcli_db_file) = @_;
	$ENV{"SYMCLI_DB_FILE"} = "$symcli_db_file";
	VCSAG_LOGDBG_MSG(2, "Executing command - symcfg discover" );
	`$symcfg discover`;
}
sub validate_db_attr {
	my ($symcli_db_file) = @_;
	if (($symcli_db_file !~ /\.bin$/) || !( -e $symcli_db_file)) {
		VCSAG_LOGDBG_MSG(1, "Invalid SymcliDbFile attribute value specified for srdf resource or Symcli db file not found" );
		return 1;
	}
	return 0;
}
1;
